#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
import io
import os
import json
from parlai.utils.misc import msg_to_str

rand = random.Random(42)


# TODO Can we grab these from somewhere in the built content?
CURRENT_HOBBOT_MODELS = [
    'orig_light_poly',
    'orig_light_poly_boring',
    'reddit_polyranker',
    'sw12_bigmlm_smallcode5',
    'sw17_bigmlm_smallcode5',
    'light_large_with_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_and_gen_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_large_with_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_large_with_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_neg?block_ngrams=3,boring_alpha=-10,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_and_gen_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=-10,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_large_with_wild_and_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_large_with_wild_and_neg?block_ngrams=3,boring_alpha=-10,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=50,normalized=none,to_speaker_bonus=50',
    'light_segment_gen_small?block_ngrams=3,boring_alpha=0,fixed_candidates_path=filtered_light_cands_segmented.txt,normalized=none',
    'light_small_with_wild_and_neg?block_ngrams=0,boring_alpha=-50,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'light_small_with_wild_and_neg?block_ngrams=0,boring_alpha=-50,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'light_small_with_wild_and_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_small_with_wild_and_neg?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'light_small_with_wild_and_neg?block_ngrams=3,boring_alpha=-50,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'orig_light_poly?block_ngrams=0,boring_alpha=-50,encode_candidate_vecs=True,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'orig_light_poly?block_ngrams=0,boring_alpha=-50,encode_candidate_vecs=True,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'orig_light_poly?block_ngrams=0,boring_alpha=-50,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'orig_light_poly?block_ngrams=0,boring_alpha=-50,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=20,normalized=none,to_speaker_bonus=20',
    'orig_light_poly?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_and_wild_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
    'orig_light_poly?block_ngrams=0,boring_alpha=0,fixed_candidates_path=filtered_light_cands.txt,from_speaker_bonus=0,normalized=none,to_speaker_bonus=0',
]


# locations in the LIGHT test unseen set.
unseen = [
    "the lords treasury",
    "mess hall of the servants",
    "grand hall of the city",
    "the endless strait",
    "frozen outpost",
    "inside of a igloo",
    "eating area of cave",
    "research laboratory",
    "the fountain of youth",
    "the dark kings throne room",
    "the cabin of an ogre",
    "lower dungeon",
    "reaper's lair",
    "stonehenge",
    "toad hollow",
    "the dungeon",
    "geothermal valley",
    "cloud tavern",
    "harp store",
    "entrance to a cave",
    "ice cave",
    "entry to icy fortress",
    "swamp",
    "faeries den",
    "a creature cave",
    "cavern",
    "torture chanber",
    "the devils castle",
    "whipping chamber",
    "lava lake",
    "ghost monk quarters",
    "astral plane",
    "bridge in the forest",
    "fighting arena inside tourney grounds",
    "tool shed",
    "farmhouse",
    "dragon lift",
    "cloud nine bar",
    "observatory",
    "wizards dragon egg incubatory",
    "the moon observatory",
    "main square of the city in the clouds",
    "hidden cave",
    "living room",
    "the abyss",
    "den",
    "festival room",
    "the royal tent",
    "large ice cave",
    "secret magician's workshop",
    "wizard's lair",
    "echo hall",
    "inside of a troll cave",
    "devil's den",
    "withered gardens",
    "the pit of despair",
    "lord's castle",
    "dungeon",
    "viewing room",
    "goddess hollow",
    "haunted basement",
    "magic room",
    "the glade of mysteries",
    "underwater marketplace",
    "research lab",
    "dry dock room",
    "underwater aquapolis",
    "the marketplace",
    "neptune's throne room",
    "the thrones of death",
    "elf's room",
]

# Score cutoff for test and valid sets
LIGHT_TEST_CUTOFF = 9


def use_feat(opt, field, ttype):
    if 'all' in opt.get(field) or opt.get(field) == ttype:
        return True
    else:
        return False


mx = 0


def fix_labels(act, opt):
    labels = act.get('labels', act.get('eval_labels'))
    clip = int(opt.get('light_use_clip_cands', 1000))
    while len(act['label_candidates']) >= clip:
        act['label_candidates'].pop()
    is_label_cand = {}
    is_label_cand[labels] = False
    for c in act['label_candidates']:
        if c in is_label_cand:
            is_label_cand[c] = True
    for l, has in is_label_cand.items():
        if has is False:
            print("***ADDING A LABEL CAND****")
            act['label_candidates'].append(l)


def get_no_affordance_actions(in_room, carrying, other_carrying, other_name):
    all_actions = []
    # Drop "a" prefix from all
    in_room = [i[2:] for i in in_room]
    carrying = [i[2:] for i in carrying]
    other_carrying = [i[2:] for i in other_carrying]
    for obj in other_carrying:
        all_actions.append('steal {} from {}'.format(obj, other_name))
    for obj in in_room:
        all_actions.append('get {}'.format(obj))
    for obj in carrying:
        for f in ['drop {}', 'wield {}', 'wear {}', 'remove {}', 'eat {}', 'drink {}']:
            all_actions.append(f.format(obj))
        all_actions.append('give {} to {}'.format(obj, other_name))
        for obj2 in in_room:
            all_actions.append('put {} in {}'.format(obj, obj2))
        for obj2 in carrying:
            if obj2 == obj:
                continue
            all_actions.append('put {} in {}'.format(obj, obj2))
    return list(set(all_actions))


def gen_no_affordance_actions_for_dialogue(d):
    characters = [d['agents'][0]['name'], d['agents'][1]['name']]
    other_char = 1
    no_affordance_actions = []
    last_carry = []
    for idx in range(len(d['speech'])):
        char = characters[other_char]
        curr_carry = d['carrying'][idx] + d['wearing'][idx] + d['wielding'][idx]
        no_affordance_actions_turn = get_no_affordance_actions(
            d['room_objects'][idx], curr_carry, last_carry, char
        )
        no_affordance_actions_turn += d['available_actions'][idx]
        no_affordance_actions.append(list(set(no_affordance_actions_turn)))
        last_carry = curr_carry
        other_char = 1 - other_char
    d['no_affordance_actions'] = no_affordance_actions


def write_dialog(opt, fw, d, label_type, split):
    l = len(d['speech'])
    msgs = []
    text = ''
    score = d['score']
    model_name = d['model_name']
    did_continue = d['did_continue']

    d['which'] = []
    for i in range(0, len(d['emote'])):
        lab = 'none'
        if d['emote'][i] is not None:
            lab = 'emote'
        if d['action'][i] is not None:
            lab = 'action'
        d['which'].append(lab)

    if opt.get('light_use_taskname', True):
        text = '_task_' + label_type + '\n'
    if opt['light_use_setting']:
        text += (
            '_setting_name '
            + d['setting']['name']
            + ", "
            + d['setting']['category']
            + '\n'
        )
        text += '_setting_desc ' + d['setting']['description'] + '\n'
    if opt['light_use_person_names']:
        text += '_partner_name ' + d['partner_agent']['name'] + '\n'
        text += '_self_name ' + d['self_agent']['name'] + '\n'
    if use_feat(opt, 'light_use_persona', 'self'):
        text += '_self_persona ' + d['self_agent']['persona'] + '\n'
    if use_feat(opt, 'light_use_persona', 'other'):
        text += '_other_persona ' + d['partner_agent']['persona'] + '\n'
    for i in range(0, l, 2):
        if i < l - 1:
            if (
                use_feat(opt, 'light_use_speech', 'partner')
                and d['speech'][i] is not None
            ):
                if opt['light_use_speech_prefix']:
                    text += '_partner_say '
                elif opt['light_use_person_names_prefix']:
                    text += f"*{d['partner_agent']['name']}*: "
                text += str(d['speech'][i]) + '\n'
            if (
                use_feat(opt, 'light_use_action', 'partner')
                and d['action'][i] is not None
            ):
                text += '_partner_act ' + str(d['action'][i]) + '\n'
            if (
                use_feat(opt, 'light_use_emote', 'partner')
                and d['emote'][i] is not None
            ):
                text += '_partner_emote ' + str(d['emote'][i]) + '\n'
            if opt.get('light_use_repeat') == 'self_last':
                if i > 0:
                    text = str(d['speech'][i - 1])
                else:
                    text = 'nothing'
            if opt.get('light_use_repeat') == 'partner_last':
                text = str(d['speech'][i])
            if opt.get('light_use_repeat') == 'both_last':
                text = ''
                if i > 0:
                    text += str(d['speech'][i - 1]) + ' '
                text += str(d['speech'][i])

            label = d[label_type][i + 1]
            if opt['light_use_person_names_prefix']:
                label = f"*{d['self_agent']['name']}*: {label}"
            used_current = False
            shown = {}
            if (
                use_feat(opt, 'light_use_current_self_output', 'speech')
                and label_type != 'speech'
                and d['speech'][i + 1] is not None
            ):
                if 'remove' not in opt['light_use_current_self_output']:
                    if opt['light_use_speech_prefix']:
                        text += '_self_say '
                    text += str(d['speech'][i + 1]) + '\n'
                    shown['speech'] = True
                used_current = True
            if (
                use_feat(opt, 'light_use_current_self_output', 'all')
                and label_type != 'action'
                and use_feat(opt, 'light_use_action', 'self')
                and d['action'][i + 1] is not None
            ):
                if 'remove' not in opt['light_use_current_self_output']:
                    text += '_self_act ' + str(d['action'][i + 1]) + '\n'
                    shown['action'] = True
                used_current = True
            if (
                use_feat(opt, 'light_use_current_self_output', 'all')
                and label_type != 'emote'
                and use_feat(opt, 'light_use_emote', 'self')
                and d['emote'][i + 1] is not None
            ):
                if 'remove' not in opt['light_use_current_self_output']:
                    text += '_self_emote ' + str(d['emote'][i + 1]) + '\n'
                    shown['emote'] = True
                used_current = True
            if (
                'all_filtered' in opt['light_use_current_self_output']
                and used_current is False
            ):
                label = None
            if label is not None:
                msg = {}
                msg['text'] = text
                msg['labels'] = label
                msg['model_name'] = model_name
                msg['did_continue'] = did_continue
                msg['score'] = score
                add_negs(
                    msg,
                    d,
                    i + 1,
                    label_type,
                    split,
                    int(opt.get('light_use_cands', 100)),
                    opt.get('light_use_affordances', True),
                )
                msgs.append(msg)
                text = ''
            if (
                use_feat(opt, 'light_use_speech', 'self')
                and d['speech'][i + 1] is not None
                and ('speech' not in shown)
            ):
                if opt['light_use_speech_prefix']:
                    text += '_self_say '
                elif opt['light_use_person_names_prefix']:
                    text += f"*{d['self_agent']['name']}*: "
                text += str(d['speech'][i + 1]) + '\n'
            if (
                use_feat(opt, 'light_use_action', 'self')
                and d['action'][i + 1] is not None
                and ('action' not in shown)
            ):
                text += '_self_act ' + str(d['action'][i + 1]) + '\n'
            if (
                use_feat(opt, 'light_use_emote', 'self')
                and d['emote'][i + 1] is not None
                and ('emote' not in shown)
            ):
                text += '_self_emote ' + str(d['emote'][i + 1]) + '\n'
    if len(msgs) > 0:
        msgs[-1]['episode_done'] = True
        for m in msgs:
            m['text'] = m['text'].rstrip('\n')
            # print(m.replace('\n', '\\n'))
            fix_labels(m, opt)
            global mx
            mx = max(len(m['label_candidates']), mx)
            # print(mx)
            txt = msg_to_str(m)
            fw.write(txt + '\n')


def write_alldata(opt, chats, dpath, ltype, split):
    # for now train, valid and test will all be identical
    fname = os.path.join(dpath, ltype + "_" + split + ".txt")
    fw_tst = io.open(fname, 'w')
    for chat in chats:
        chat = chat.copy()
        try:
            chat['self_agent'] = {
                'name': chat['human_persona']['name'],
                'persona': chat['human_persona']['persona'],
            }
            chat['partner_agent'] = {
                'name': chat['bot_persona']['name'],
                'persona': chat['bot_persona']['persona'],
            }
        except Exception as e:
            print(chat)
            raise (e)
        chat['setting'] = {
            'name': chat['location']['name'],
            'category': "Somewhere",  # TODO pull from DB?
            'description': chat['location']['description'],
        }
        turns = chat['dialogue']
        fin_turns = [t['text'] for t in turns]
        chat['speech'] = [None] + fin_turns
        chat['action'] = [None] * len(chat['speech'])
        chat['emote'] = [None] * len(chat['speech'])
        if opt['light_use_unseen_test']:
            setting = chat['location']['name'].lower()
            if setting not in unseen:
                #   import pdb; pdb.set_trace()
                write_dialog(opt, fw_tst, chat, ltype, split)
        else:
            write_dialog(opt, fw_tst, chat, ltype, split)
    fw_tst.close()


def add_negs(msg, d, ind, label_type, split, num_cands, use_affordances):
    if label_type == 'emote':
        msg['label_candidates'] = cands['emote']
    if label_type == 'which':
        msg['label_candidates'] = cands['which']
    if label_type == 'action':
        if use_affordances:
            msg['label_candidates'] = d['available_actions'][ind]
        else:
            msg['label_candidates'] = d['no_affordance_actions'][ind]
    if label_type == 'speech':
        cnt = 0
        label = msg['labels']
        negs = []
        # negs.append(label)
        used = {}
        # find 99 random negs that are != gold label
        while True:
            ind = rand.randint(0, len(cands['speech']) - 1)
            txt = cands['speech'][ind]
            if txt != label and ind not in used:
                negs.append(txt)
                used[ind] = True
                cnt += 1
                if cnt == num_cands - 1:
                    break
        # insert label in a random position
        negs.insert(rand.randrange(len(negs) + 1), label)
        msg['label_candidates'] = negs


cands = {}  # candidates


def write_out_candidates(chats, dpath, dtype):
    cands['speech'] = []
    for chat in chats:
        dialogue_turns = chat['dialogue']
        pos_cands = [t['text'] for t in dialogue_turns if t['id'] == 'human']
        cands['speech'] += pos_cands
    fw = io.open(os.path.join(dpath, "speech_" + dtype + "_cands.txt"), 'w')
    for k in cands['speech']:
        fw.write(k + "\n")
    fw.close


def read_dump(dump_path):
    with open(dump_path, 'r') as chat_file:
        chats = json.load(chat_file)

    if isinstance(chats, dict):
        headers = chats['header']
        rows = chats['rows']
        chats = []
        for row in rows:
            chat = {}
            for idx, header in enumerate(headers):
                chat[header] = row[idx]
            chats.append(chat)
    return chats


def remove_cand_path(in_path):
    fcp_split = in_path.split('fixed_candidates_path')
    pre_fixed_cands = fcp_split[0]
    post_fixed_cands = fcp_split[1].split('/')[-1]
    return pre_fixed_cands + "fixed_candidates_path=" + post_fixed_cands


def select_location_from_rough_string_match(location1, location2, chat):
    loc1_score = 0.0
    loc2_score = 0.0
    loc1_lower = location1.lower()
    loc2_lower = location2.lower()
    for turn in chat:
        for word in turn.lower().split(' '):
            if len(word) < 5:
                continue
            if word in loc1_lower:
                loc1_score += 1
            if word in loc2_lower:
                loc2_score += 1
    loc1_score /= len(loc1_lower.split(' '))
    loc2_score /= len(loc2_lower.split(' '))
    if loc1_score > loc2_score:
        return location1
    else:
        return location2


def format_chats_orig(chats):
    """
    Original hobbot chat format stored the dialogues as newline separated text, and the
    bot and human personas as strings.

    The choice and location appear in the dialogue history as strings.
    """
    chats = [
        c
        for c in chats
        if c['dialogue'] is not None and len(c['dialogue'].split('\\n')) > 1
    ]

    converted_chats = []
    for c in chats:
        try:
            # set up personas
            old_human_persona = c['human_persona'].split('\\n')
            c['human_persona'] = {
                'id': None,
                'name': old_human_persona[0],
                'persona': old_human_persona[1],
                'db_id': None,
            }
            old_bot_persona = c['bot_persona'].split('\\n')
            c['bot_persona'] = {
                'id': None,
                'name': old_bot_persona[0],
                'persona': old_bot_persona[1],
                'db_id': None,
            }

            old_dialogue_turns = c['dialogue'].split('\\n')
            if old_dialogue_turns[0].startswith("Human:"):
                # The very first 125 chats we collected didn't add the
                # specific location, so there are two possibilities.
                # We select the possible location with the
                # most word overlap with the chat.
                use_location = select_location_from_rough_string_match(
                    old_human_persona[2], old_bot_persona[2], old_dialogue_turns
                )
                old_dialogue_turns = [use_location] + old_dialogue_turns
            # extract location:
            location_turn_split = old_dialogue_turns[0].split(', ')
            c['location'] = {
                'id': None,
                'name': location_turn_split[0],
                'description': ', '.join(location_turn_split[1:]),
                'db_id': None,
            }

            # parse through dialogue
            new_dialogue = []
            c['choice'] = None
            BOT_TAG = 'Bot: '
            CHOICE_TAG = 'CHOICE: '
            HUMAN_TAG = 'Human: '
            c['score'] = int(c['score'])
            score_per_turn = c['score'] / max(1, len(old_dialogue_turns) / 2)
            for turn in old_dialogue_turns[1:]:
                if turn.startswith(BOT_TAG):
                    new_dialogue.append(
                        {'id': 'bot', 'type': 'speech', 'text': turn[len(BOT_TAG) :]}
                    )
                elif turn.startswith(HUMAN_TAG):
                    new_dialogue.append(
                        {
                            'id': 'human',
                            'type': 'speech',
                            'text': turn[len(HUMAN_TAG) :],
                            'score': score_per_turn,
                        }
                    )
                elif turn.startswith(CHOICE_TAG):
                    c['choice'] = {
                        'id': 'human',
                        'type': 'choice',
                        'text': turn[len(CHOICE_TAG) :],
                    }
            c['dialogue'] = new_dialogue
            c['is_complete'] = c['choice'] is not None
            c['did_continue'] = (
                c['choice'] is not None and c['choice']['text'] != 'EXIT'
            )
            converted_chats.append(c)
        except Exception:
            if c['ds'] == '2020-06-23':
                # We expect some chats from this day to have
                # broken formatting
                continue
            else:
                raise

    return converted_chats


def format_chats_june_2020(chats):
    """
    June 2020 format involves storing location and choice in the dialogue, but complete
    human and bot personas as objects that need to be json loaded.

    Model name also contains the full fixed candidates path, when we're only truly
    interested in the filename.
    """
    for c in chats:
        c['dialogue'] = json.loads(c['dialogue'].replace('\\\\', '\\'))
        c['human_persona'] = json.loads(c['human_persona'].replace('\\\\"', '\\"'))
        c['bot_persona'] = json.loads(c['bot_persona'].replace('\\\\"', '\\"'))
        c['model_name'] = remove_cand_path(c['model_name'].replace('\\\\"', '\\"'))
    chats = [c for c in chats if c['dialogue'] is not None and len(c['dialogue']) > 1]

    # Parse the choice and location out from the dialogue
    for c in chats:
        last_chat = c['dialogue'][-1]
        if last_chat['type'] == 'choice':
            c['choice'] = last_chat
            c['dialogue'] = c['dialogue'][:-1]
        else:
            c['choice'] = None

        c['location'] = c['dialogue'][0]
        c['dialogue'] = c['dialogue'][1:]
        c['score'] = int(c['score'])
        c['is_complete'] = c['choice'] is not None
        c['did_continue'] = c['choice'] is not None and c['choice']['text'] != 'EXIT'

        for t in c['dialogue']:
            if len(t['text'].strip()) == 0:
                t['text'] = '__SILENCE__'

    return chats


def format_chats(chat_list):
    if chat_list[0]['ds'] < '2020-06-24':
        return format_chats_orig(chat_list)
    else:
        return format_chats_june_2020(chat_list)


def read_dumps(dump_dir):
    dump_files = os.listdir(dump_dir)
    dump_files.sort()
    all_chats = []
    for dump_fn in dump_files:
        if not dump_fn.endswith('.json'):
            continue
        read_chats = read_dump(os.path.join(dump_dir, dump_fn))
        all_chats += format_chats(read_chats)
    return all_chats


def filter_chats(chats, return_flagged=False):
    chats = [c for c in chats if c['dialogue'] is not None and len(c['dialogue']) > 0]
    return [c for c in chats if (c['flagged_messages'] is not None) == return_flagged]


def remove_under_score_cutoff(chats, score_cutoff):
    return [c for c in chats if c['score'] >= score_cutoff]


def remove_over_score_cutoff(chats, score_cutoff):
    return [c for c in chats if c['score'] <= score_cutoff]


def remove_except_score_cutoff(chats, score_cutoff):
    return [c for c in chats if c['score'] == score_cutoff]


def filter_by_model(chats, model_name):
    return [c for c in chats if c['model_name'] == model_name]


def filter_by_chat_date(chats, chat_date):
    return [c for c in chats if c['ds'] <= chat_date]


def filter_by_continue_type(chats, continue_type):
    desired_continue_status = continue_type == 'continue'
    return [c for c in chats if c['did_continue'] == desired_continue_status]


def build_from_dump(opt, write_path, dump_dir):
    # set up chats
    chats = read_dumps(dump_dir)
    filtered_chats = filter_chats(chats)
    valid_choice_chats = filtered_chats[:10000]
    remaining_chats = filtered_chats[10000:]
    rand = random.Random(3)
    rand.shuffle(valid_choice_chats)

    # Construct a validation and test set with a preset cutoff
    above_test_cutoff = remove_under_score_cutoff(
        valid_choice_chats, score_cutoff=LIGHT_TEST_CUTOFF
    )
    below_test_cutoff = [
        c for c in valid_choice_chats if c['score'] < LIGHT_TEST_CUTOFF
    ]

    test_chats = above_test_cutoff[:1000]
    valid_chats = above_test_cutoff[1000:1500]
    remaining_chats += above_test_cutoff[1500:]

    trainable_chats = remaining_chats + below_test_cutoff
    if opt['light_use_hard_score_cutoff']:
        train_chats = remove_except_score_cutoff(
            trainable_chats, score_cutoff=opt['light_use_score_cutoff']
        )
    elif opt['light_use_max_score_cutoff'] > 0:
        train_chats = remove_over_score_cutoff(
            trainable_chats, score_cutoff=opt['light_use_max_score_cutoff']
        )
    else:
        train_chats = remove_under_score_cutoff(
            trainable_chats, score_cutoff=opt['light_use_score_cutoff']
        )

    if opt.get('light_model_name'):
        model_name = opt['light_model_name'].split('+')
        assert all(m in CURRENT_HOBBOT_MODELS for m in model_name)
        train_chats = sum([filter_by_model(train_chats, m) for m in model_name], [])

    if opt['light_use_continue_type'] != 'all':
        train_chats = filter_by_continue_type(
            train_chats, opt['light_use_continue_type']
        )

    if opt.get('light_use_date_cutoff'):
        date_cutoff = opt['light_use_date_cutoff']
        train_chats = filter_by_chat_date(train_chats, date_cutoff)

    write_out_candidates(test_chats, write_path, 'test')
    write_alldata(opt, test_chats, write_path, 'speech', 'test')

    write_out_candidates(valid_chats, write_path, 'valid')
    write_alldata(opt, valid_chats, write_path, 'speech', 'valid')

    write_out_candidates(train_chats, write_path, 'train')
    write_alldata(opt, train_chats, write_path, 'speech', 'train')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Config Utils.
"""

import yaml
from collections import namedtuple

WorldConfig = namedtuple(
    "WorldConfig",
    [
        "world_name",
        "onboarding_name",
        "task_name",
        "max_time_in_pool",
        "agents_required",
        "backup_task",
    ],
)


def parse_configuration_file(config_path):
    """
    Read the config file for an experiment to get ParlAI settings.

    :param config_path:
        path to config

    :return:
        parsed configuration dictionary
    """
    result = {}
    result["configs"] = {}
    with open(config_path) as f:
        cfg = yaml.load(f.read(), Loader=yaml.SafeLoader)
        # get world path
        result["world_path"] = cfg.get("world_module")
        if not result["world_path"]:
            raise ValueError("Did not specify world module")
        result["overworld"] = cfg.get("overworld")
        if not result["overworld"]:
            raise ValueError("Did not specify overworld")
        result["max_workers"] = cfg.get("max_workers")
        if not result["max_workers"]:
            raise ValueError("Did not specify max_workers")
        result["task_name"] = cfg.get("task_name")
        if not result["task_name"]:
            raise ValueError("Did not specify task name")
        task_world = cfg.get("tasks")
        if task_world is None or len(task_world) == 0:
            raise ValueError("task not in config file")
        # get task file
        for task_name, configuration in task_world.items():
            if "task_world" not in configuration:
                raise ValueError("{} does not specify a task".format(task_name))
            result["configs"][task_name] = WorldConfig(
                world_name=task_name,
                onboarding_name=configuration.get("onboard_world"),
                task_name=configuration.get("task_world"),
                max_time_in_pool=configuration.get("timeout") or 300,
                agents_required=configuration.get("agents_required") or 1,
                backup_task=configuration.get("backup_task"),
            )
        # get world options, additional args
        result["world_opt"] = cfg.get("opt", {})
        result["additional_args"] = cfg.get("additional_args", {})

    return result



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/COCO-IMG/test2015.zip',
        'test2015.zip',
        '52d45179bbe4fcc41ba16550b3df532fc9d0a0084b2afaeb6a3ae396032aaf14',
    ),
    DownloadableFile(
        'http://images.cocodataset.org/annotations/image_info_test2015.zip',
        'image_info_test2015.zip',
        'cf400242f8497257fb8a3e369bc766491f4a7e42625fb3d72555504e9a8c3b18',
    ),
]


def buildImage(opt):
    dpath = os.path.join(opt['datapath'], 'COCO-IMG-2015')
    version = '1'

    if not build_data.built(dpath, version_string=version):
        print('[building image data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES[:1]:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)


def build(opt):
    dpath = os.path.join(opt['datapath'], 'COCO_2015_Caption')
    version = None

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES[1:]:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.build_data import modelzoo_path
from parlai.utils.io import PathManager


def set_defaults(opt):
    init_model = None
    # check first for 'init_model' for loading model from file
    if opt.get('init_model') and PathManager.exists(opt['init_model']):
        init_model = opt['init_model']
    # next check for 'model_file', this would override init_model
    if opt.get('model_file') and PathManager.exists(opt['model_file']):
        init_model = opt['model_file']

    if init_model is None:
        # Embeddings options
        opt['embedding_file'] = modelzoo_path(
            opt.get('datapath'), opt['embedding_file']
        )
        if opt.get('embedding_file'):
            if not PathManager.exists(opt['embedding_file']):
                raise IOError('No such file: %s' % opt['embedding_file'])
            with PathManager.open(opt['embedding_file']) as f:
                dim = len(f.readline().strip().split(' ')) - 1
                if dim == 1:
                    # first line was a dud
                    dim = len(f.readline().strip().split(' ')) - 1
            opt['embedding_dim'] = dim
        elif not opt.get('embedding_dim'):
            raise RuntimeError(
                ('Either embedding_file or embedding_dim ' 'needs to be specified.')
            )

        # Make sure tune_partial and fix_embeddings are consistent
        if opt['tune_partial'] > 0 and opt['fix_embeddings']:
            print('Setting fix_embeddings to False as tune_partial > 0.')
            opt['fix_embeddings'] = False

        # Make sure fix_embeddings and embedding_file are consistent
        if opt['fix_embeddings'] and not opt.get('embedding_file'):
            print('Setting fix_embeddings to False as embeddings are random.')
            opt['fix_embeddings'] = False


def override_args(opt, override_opt):
    # Major model args are reset to the values in override_opt.
    # Non-architecture args (like dropout) are kept.
    args = set(
        [
            'embedding_file',
            'embedding_dim',
            'hidden_size',
            'doc_layers',
            'question_layers',
            'rnn_type',
            'optimizer',
            'concat_rnn_layers',
            'question_merge',
            'use_qemb',
            'use_in_question',
            'use_tf',
            'vocab_size',
            'num_features',
            'use_time',
        ]
    )
    for k, v in override_opt.items():
        if k in args:
            opt[k] = v



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import hashlib
import os
from typing import Iterable, List, Tuple, Union

import pandas as pd
from pytest_regressions.data_regression import DataRegressionFixture
from pytest_regressions.dataframe_regression import DataFrameRegressionFixture
from pytest_regressions.file_regression import FileRegressionFixture

from parlai.crowdsourcing.utils.tests import AbstractOneTurnCrowdsourcingTest


TASK_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


def get_hashed_combo_path(
    root_dir: str,
    subdir: str,
    task: str,
    combos: Iterable[Union[List[str], Tuple[str, str]]],
) -> str:
    """
    Return a unique path for the given combinations of models.

    :param root_dir: root save directory
    :param subdir: immediate subdirectory of root_dir
    :param task: the ParlAI task being considered
    :param combos: the combinations of models being compared
    """

    # Sort the names in each combo, as well as the overall combos
    sorted_combos = []
    for combo in combos:
        assert len(combo) == 2
        sorted_combos.append(tuple(sorted(combo)))
    sorted_combos = sorted(sorted_combos)

    os.makedirs(os.path.join(root_dir, subdir), exist_ok=True)
    path = os.path.join(
        root_dir,
        subdir,
        hashlib.sha1(
            '___and___'.join(
                [f"{m1}vs{m2}.{task.replace(':', '_')}" for m1, m2 in sorted_combos]
            ).encode('utf-8')
        ).hexdigest()[:10],
    )
    return path


class AbstractFastAcuteTest(AbstractOneTurnCrowdsourcingTest):
    """
    Abstract test class for testing Fast ACUTE code.
    """

    TASK_DIRECTORY = TASK_DIRECTORY
    MODELS = ['model1', 'model2']
    MODEL_STRING = ','.join(MODELS)
    TASK_DATA = {
        "final_data": [
            {"speakerChoice": "human_as_model", "textReason": "Makes more sense"},
            {"speakerChoice": "model1", "textReason": "Makes more sense"},
            {"speakerChoice": "model2", "textReason": "Makes more sense"},
            {"speakerChoice": "model1", "textReason": "Makes more sense"},
            {"speakerChoice": "model2", "textReason": "Makes more sense"},
        ]
    }

    def _get_common_overrides(self, root_dir: str) -> List[str]:
        """
        Return overrides for all subclassed Fast ACUTE test code.
        """
        # TODO: clean this up when Hydra has support for recursive defaults
        return [
            '+mephisto.blueprint.acute_eval_type=engaging',
            'mephisto.blueprint.block_on_onboarding_fail=False',
            '+mephisto.blueprint.matchups_per_pair=60',
            '+mephisto.blueprint.num_self_chats=5',
            f'+mephisto.blueprint.onboarding_path={self.TASK_DIRECTORY}/task_config/onboarding.json',
            f'+mephisto.blueprint.root_dir={root_dir}',
            '+mephisto.blueprint.sufficient_matchups_multiplier=2',
            '+mephisto.blueprint.task=blended_skill_talk',
            'mephisto.task.task_name=acute_eval_test',
        ]

    def test_agent_state(self, setup_teardown, data_regression: DataRegressionFixture):
        outputs = setup_teardown
        self._check_agent_state(state=outputs['state'], data_regression=data_regression)

    def test_all_convo_pairs_txt(
        self, setup_teardown, file_regression: FileRegressionFixture
    ):
        outputs = setup_teardown
        self._check_file_contents(
            results_folder=outputs['results_folder'],
            file_suffix='all_convo_pairs.txt',
            file_regression=file_regression,
        )

    def test_all_html(self, setup_teardown, file_regression: FileRegressionFixture):
        outputs = setup_teardown
        self._check_file_contents(
            results_folder=outputs['results_folder'],
            file_suffix='all.html',
            file_regression=file_regression,
        )

    def test_full_csv(
        self, setup_teardown, dataframe_regression: DataFrameRegressionFixture
    ):
        outputs = setup_teardown
        self._check_dataframe(
            results_folder=outputs['results_folder'],
            file_suffix='full.csv',
            dataframe_regression=dataframe_regression,
        )

    def test_grid_csv(
        self, setup_teardown, dataframe_regression: DataFrameRegressionFixture
    ):
        outputs = setup_teardown
        self._check_dataframe(
            results_folder=outputs['results_folder'],
            file_suffix='grid.csv',
            dataframe_regression=dataframe_regression,
        )

    def test_grid_winners_as_rows_csv(
        self, setup_teardown, dataframe_regression: DataFrameRegressionFixture
    ):
        outputs = setup_teardown
        self._check_dataframe(
            results_folder=outputs['results_folder'],
            file_suffix='grid.winners_as_rows.csv',
            dataframe_regression=dataframe_regression,
        )

    def test_ratings_per_worker_csv(
        self, setup_teardown, dataframe_regression: DataFrameRegressionFixture
    ):
        outputs = setup_teardown
        self._check_dataframe(
            results_folder=outputs['results_folder'],
            file_suffix='ratings_per_worker.csv',
            dataframe_regression=dataframe_regression,
        )

    def test_reason_html(self, setup_teardown, file_regression: FileRegressionFixture):
        outputs = setup_teardown
        self._check_file_contents(
            results_folder=outputs['results_folder'],
            file_suffix='reason.html',
            file_regression=file_regression,
        )

    def test_significance_csv(
        self, setup_teardown, dataframe_regression: DataFrameRegressionFixture
    ):
        outputs = setup_teardown
        self._check_dataframe(
            results_folder=outputs['results_folder'],
            file_suffix='significance.csv',
            dataframe_regression=dataframe_regression,
        )

    def _check_dataframe(
        self,
        results_folder: str,
        file_suffix: str,
        dataframe_regression: DataFrameRegressionFixture,
    ):
        file_path = self._get_matching_file_path(
            results_folder=results_folder, file_suffix=file_suffix
        )
        df = pd.read_csv(file_path)
        dataframe_regression.check(data_frame=df)

    def _check_file_contents(
        self,
        results_folder: str,
        file_suffix: str,
        file_regression: FileRegressionFixture,
    ):
        file_path = self._get_matching_file_path(
            results_folder=results_folder, file_suffix=file_suffix
        )
        with open(file_path) as f:
            contents = f.read()
        file_regression.check(contents=contents)

    def _get_matching_file_path(self, results_folder: str, file_suffix: str) -> str:
        matching_files = [
            obj for obj in os.listdir(results_folder) if obj.endswith(file_suffix)
        ]
        assert len(matching_files) == 1
        return os.path.join(results_folder, matching_files[0])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Utility methods for dealing with torch code.
"""

import os
from typing import Union, Optional, Tuple, Any, List, Sized, TypeVar
import itertools
from collections import namedtuple
import parlai.utils.logging as logging
import parlai.utils.io as io_utils


try:
    import torch
except ImportError:
    raise ImportError('Parlai requires pytorch. Go to http://pytorch.org to install.')

import torch.optim

"""Near infinity, useful as a large penalty for scoring when inf is bad."""
NEAR_INF = 1e20
NEAR_INF_FP16 = 65504

# according to the tensor cores documentation from nvidia, the matmuls in fp16
# must all be multiples of 8 in order to get the speedup from fp16. We set this
# as a constant here for clarity and convenience.  See
# https://devblogs.nvidia.com/programming-tensor-cores-cuda-9/ for more
# information.
FP16_PAD_SIZE = 8


def neginf(dtype: torch.dtype) -> float:
    """
    Return a representable finite number near -inf for a dtype.
    """
    if dtype is torch.float16:
        return -NEAR_INF_FP16
    else:
        return -NEAR_INF


def atomic_save(state_dict: Any, path: str) -> None:
    """
    Like torch.save, but atomic.

    Useful for preventing trouble coming from being pre-empted or killed while writing
    to disk. Works by writing to a temporary file, and then renaming the file to the
    final name.
    """

    if io_utils.USE_ATOMIC_TORCH_SAVE:
        with open(path + ".tmp", "wb") as f:
            torch.save(state_dict, f)
        os.replace(path + ".tmp", path)
    else:
        with io_utils.PathManager.open(path, "wb") as f:
            torch.save(state_dict, f)


def padded_tensor(
    items: List[Union[List[int], torch.LongTensor]],
    pad_idx: int = 0,
    use_cuda: bool = False,
    left_padded: bool = False,
    max_len: Optional[int] = None,
    fp16friendly: bool = False,
    device: int = -1,
) -> Tuple[torch.LongTensor, List[int]]:
    """
    Create a padded matrix from an uneven list of lists.

    Returns (padded, lengths), where padded is the padded matrix, and lengths
    is a list containing the lengths of each row.

    Matrix is right-padded (filled to the right) by default, but can be
    left padded if the flag is set to True.

    Matrix can also be placed on cuda automatically.

    :param list[iter[int]] items: List of items
    :param bool sort: If True, orders by the length
    :param int pad_idx: the value to use for padding
    :param bool use_cuda: if true, places `padded` on GPU
    :param bool left_padded:
    :param int max_len: if None, the max length is the maximum item length
    :param bool fp16friendly: if True, pads the time dimension to be a multiple of 4.
    :param int device: GPU device.

    :returns: (padded, lengths) tuple
    :rtype: (Tensor[int64], list[int])
    """

    # number of items
    n = len(items)
    # length of each item
    lens: List[int] = [len(item) for item in items]  # type: ignore
    # max in time dimension
    t = max(lens) if max_len is None else max_len

    # if input tensors are empty, we should expand to nulls
    t = max(t, 1)

    if fp16friendly and (t % FP16_PAD_SIZE != 0):
        # pad to be fp16 friendly
        t += FP16_PAD_SIZE - (t % FP16_PAD_SIZE)

    if isinstance(items[0], torch.Tensor):
        # keep type of input tensors, they may already be cuda ones
        output = items[0].new(n, t)  # type: ignore
    else:
        output = torch.LongTensor(n, t)  # type: ignore
    output.fill_(pad_idx)

    for i, (item, length) in enumerate(zip(items, lens)):
        if length == 0:
            # skip empty items
            continue
        if not isinstance(item, torch.Tensor):
            # put non-tensors into a tensor
            item = torch.LongTensor(item)  # type: ignore
        if left_padded:
            # place at end
            output[i, t - length :] = item
        else:
            # place at beginning
            output[i, :length] = item

    if use_cuda:
        output = output.cuda()
        if device >= 0:
            output = output.to(device)
    return output, lens


def padded_3d(
    tensors: List[torch.LongTensor],
    pad_idx: int = 0,
    use_cuda: bool = False,
    dtype: Optional[torch.dtype] = torch.long,
    fp16friendly: bool = False,
):
    """
    Make 3D padded tensor for list of lists of 1D tensors or lists.

    :param tensors:
        list of lists of 1D tensors (or lists)
    :param pad_idx:
        padding to fill tensor with
    :param use_cuda:
        whether to call cuda() before returning
    :param bool fp16friendly:
        if True, pads the final dimension to be a multiple of 8.

    :returns:
        3D tensor with the maximum dimensions of the inputs
    """
    a = len(tensors)
    b = max(len(row) for row in tensors)  # type: ignore
    c = max(len(item) for row in tensors for item in row)  # type: ignore

    # pad empty tensors
    if fp16friendly and c % FP16_PAD_SIZE != 0:
        c += FP16_PAD_SIZE - (c % FP16_PAD_SIZE)
    c = max(c, 1)

    output = torch.full((a, b, c), pad_idx, dtype=dtype)

    for i, row in enumerate(tensors):
        item: Sized
        for j, item in enumerate(row):  # type: ignore
            if len(item) == 0:
                continue
            if not isinstance(item, torch.Tensor):
                item = torch.Tensor(item, dtype=dtype)
            output[i, j, : len(item)] = item

    if use_cuda:
        output = output.cuda()

    return output


def concat_without_padding(text_idx, cand_idx, use_cuda, null_idx=0):
    """
    Concatenate two right padded tensors and move padding to the right.

    For example,
        if text_idx = [[1, 2, 3, 4, 0, 0  ]]
        and cand_idx = [[5, 6, 7, 8, 0, 0 ]]:
    Then result = (tokens, segments) where
        tokens = [[1, 2, 3, 4, 5, 6, 7, 8, 0, 0, 0, 0]]
        segments = [[0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]]
    """
    assert text_idx.size(0) == cand_idx.size(0)
    assert len(text_idx.size()) == 2
    assert len(cand_idx.size()) == 2
    segments_idx = [0, 1]
    text_idx = text_idx.cpu()
    cand_idx = cand_idx.cpu()
    cand_len = cand_idx.size(1)
    concat_len = text_idx.size(1) + cand_idx.size(1)
    tokens = text_idx.new_zeros(text_idx.size(0), concat_len) + null_idx
    segments = text_idx.new_zeros(text_idx.size(0), concat_len) + null_idx
    for i in range(len(tokens)):
        non_nuls = torch.sum(text_idx[i, :] != null_idx)
        tokens[i, 0:non_nuls] = text_idx[i, 0:non_nuls]
        segments[i, 0:non_nuls] = segments_idx[0]
        tokens[i, non_nuls : non_nuls + cand_len] = cand_idx[i, :]
        segments[i, non_nuls : non_nuls + cand_len] = segments_idx[1]
    if use_cuda:
        tokens = tokens.cuda()
        segments = segments.cuda()
    return tokens, segments


def argsort(keys: List[Any], *lists: List[List[Any]], descending: bool = False):
    """
    Reorder each list in lists by the (descending) sorted order of keys.

    :param iter keys:
        Keys to order by.
    :param list[list] lists:
        Lists to reordered by keys's order.  Correctly handles lists and 1-D
        tensors.
    :param bool descending:
        Use descending order if true.

    :returns:
        The reordered items.
    """
    ind_sorted = sorted(range(len(keys)), key=lambda k: keys[k])
    if descending:
        ind_sorted = list(reversed(ind_sorted))
    output = []
    for lst in lists:
        # watch out in case we don't have torch installed
        if isinstance(lst, torch.Tensor):
            output.append(lst[ind_sorted])
        else:
            output.append([lst[i] for i in ind_sorted])
    return output


def compute_grad_norm(parameters, norm_type=2.0):
    """
    Compute norm over gradients of model parameters.

    :param parameters:
        the model parameters for gradient norm calculation. Iterable of
        Tensors or single Tensor
    :param norm_type:
        type of p-norm to use

    :returns:
        the computed gradient norm
    """
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    parameters = [p for p in parameters if p is not None and p.grad is not None]
    total_norm = 0
    for p in parameters:
        param_norm = p.grad.data.norm(norm_type)
        total_norm += param_norm.item() ** norm_type
    return total_norm ** (1.0 / norm_type)


class IdentityLayer(torch.nn.Module):
    """
    Identity layer module.

    Useful for decoder-only Torch Generator agents.
    """

    def forward(self, xs):
        """
        Identity.
        """
        return xs


def total_parameters(model: torch.nn.Module) -> int:
    """
    Count the total number of parameters in the model.

    :param model:
        the model whose parameters we wish to count.

    :return:
        total number of parameters in the model.
    """
    return sum(p.numel() for p in model.parameters())


def trainable_parameters(model: torch.nn.Module) -> int:
    """
    Count the total number of trainable parameters in the model.

    :param model:
        the model whose parameters we wish to count.

    :return:
        total number of trainable parameters in the model.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


Chunk = TypeVar('Chunk')


PipelineWorkItem = namedtuple(
    'PipelineWorkItem', ['chunk_idx', 'layer_nos', 'next_device']
)


class PipelineHelper(object):
    """
    PipelineHelper assists with implementing pipelining in model parallelism.

    For a tutorial on model parallelism, as it's implemented in parts of ParlAI,
    see https://pytorch.org/tutorials/intermediate/model_parallel_tutorial.html.

    Usage:
    >>> my_model = PipelineHelper().make_parallel(my_model)

    Note that you will need to manually implement logic which handles the
    moved layers.
    """

    def __init__(self):
        self.__device_allocations = {}
        self.num_devices = torch.cuda.device_count()
        self.devices = []
        for i in range(self.num_devices):
            d = f'cuda:{i}'
            self.devices.append(d)
            self.__device_allocations[d] = 0

    def check_compatibility(self, opt):
        """
        Check compatibility for opts.

        Really just used to raise an error message if the user mixes multiprocessing and
        model parallelism.
        """
        if opt.get('multiprocessing') and not os.environ.get('PARLAI_FORCE_MP'):
            raise RuntimeError(
                "It looks like you are trying to mix multiprocessing data "
                "parallelism (multiprocessing_train or multiprocessing_eval) "
                "with --model-parallel true. This is almost certainly a user "
                "error, and is going to result in hanging as the two methods "
                "fight for resources. Use simple `train_model` instead of "
                "`mp_train`, or add `--model-parallel false`. For more info, "
                "see https://github.com/facebookresearch/ParlAI/issues/2962."
            )

    def make_parallel(self, model: torch.nn.Module) -> torch.nn.Module:
        """
        Allocate specific layers in a model to be ModelParallel.

        Limited to only ModuleLists within the model.  Uses some heuristics to
        attempt to evenly distribute layers across GPUs, in order to balance
        memory usage. They are:

        - Assume the 0th GPU will host the optimizer, word embeddings, etc.
        - Assume activation memory is linear with the number of parameters.
        - All layers are approximately equal in size.
        """

        # first assume all layers will go onto gpu 0 as an optimizer. The
        # optimizer and embeddings are not quite as expensive as the
        # activations (which scale via batchsize), Empirically, I found this
        # heuristic works well enough. The weighting factor of 3 is more or
        # less made up.
        self.__device_allocations['cuda:0'] += trainable_parameters(model) * 3

        model.apply(self._place_modulelist)
        model._apply(self._move_rest_to_cuda0)  # type: ignore
        return model

    def _move_rest_to_cuda0(self, parameter: torch.Tensor):
        if parameter.device.type == 'cpu':
            return parameter.to('cuda:0')
        else:
            return parameter

    def _place_modulelist(self, submodule: torch.nn.Module) -> None:
        if not isinstance(submodule, torch.nn.ModuleList):
            # not a ModuleList, leave it untouched
            return
        if getattr(submodule, 'model_parallel_exempt', False):
            return

        assert isinstance(submodule, torch.nn.ModuleList)  # for typechecker
        layers = submodule

        # mark this section as MP
        layers.is_model_parallel = True  # type: ignore

        # next, let's figure out how many parameters we can assign to each GPU,
        # but not make actual assignments yet. Assignments come later because we
        # want consecutive layers to be collocated
        keyfunc = self.__device_allocations.__getitem__
        layer_assignments = {k: 0 for k in self.devices}
        for layer_no, layer in enumerate(layers):
            if layer_no == 0:
                # hard code the first layer to be 0.
                mostfree = 'cuda:0'
            else:
                # otherwise dynamic allocation
                mostfree = min(self.devices, key=keyfunc)
            # 32 is a totally arbitrary, made up number that worked in practice
            # on the large models I tested on. I believe it should be roughly
            # batch size, but this was set empirically.
            self.__device_allocations[mostfree] += trainable_parameters(layer) * 32
            # mark a layer as going to the given element
            layer_assignments[mostfree] += 1

        devices = [d for i, d in enumerate(self.devices[:]) if layer_assignments[d] > 0]
        for layer_no, layer in enumerate(layers):
            layer_gpu = devices[0]
            assert layer_assignments[layer_gpu] > 0
            logging.debug(f"Model Parallel: Assigning {layer_no} to {layer_gpu}")
            layer._mp_gpu = layer_gpu
            layers[layer_no] = layer.to(layer_gpu)
            layer_assignments[layer_gpu] -= 1
            if layer_assignments[layer_gpu] == 0:
                devices.pop(0)

    @staticmethod
    def guess_split_size(item: Chunk, num_gpus: Optional[int] = None, dim=0) -> int:
        """
        Estimate the number of chunks we should split the batch into via heuristics.
        """
        if num_gpus is None:
            num_gpus = torch.cuda.device_count()  # type: ignore

        if isinstance(item, torch.Tensor):
            if num_gpus == 1:
                # no point in chunking if we're not really doing model parallel
                return item.size(dim)
            # heuristic: use the same number of chunks as 2 * num_gpus.  this
            # isn't perfect (it ideally would be tuned differently for every model
            # and number of GPUs), but it seems to work reasonably wellenough in several
            # architectures tested.
            return max(1, item.size(dim) // int(num_gpus * 2))
        elif isinstance(item, tuple):
            return PipelineHelper.guess_split_size(item[0], num_gpus)
        elif isinstance(item, dict):
            return PipelineHelper.guess_split_size(list(item.values())[0], num_gpus)
        raise TypeError(f'Cannot determine split size for {type(item)}')

    @staticmethod
    def split(item: Chunk, split_size: Optional[int] = None, dim=0) -> List[Chunk]:
        """
        Split a tensor or group of tensors into smaller chunks of the same type.

        :param item:
            The item being split. May be a Tensor, a tuple of Tensors, or a
            dictionary mapping str -> Tensor.
        :param split_size:
            The maximum size of each output chunk. If None, we will guess using
            heuristics
        :param dim:
            The dimension to split along.
        """
        if split_size is None:
            split_size = PipelineHelper.guess_split_size(item)

        if isinstance(item, torch.Tensor):
            # base case, just split the tensor
            return list(torch.split(item, split_size, dim))
        elif isinstance(item, tuple):
            # We start with Tuple[Tensor] and we return List[Tuple[Tensor]]
            return list(zip(*(PipelineHelper.split(i, split_size, dim) for i in item)))
        elif isinstance(item, dict):
            if item == {}:
                # Terrible edge case: the empty dict. We handle by returning an
                # infinite list of empty dicts and we'll figure out its correct
                # size later. This happens for the incremental_state in
                # MultiheadAttention.
                return itertools.repeat({})  # type: ignore

            # we can't handle dicts with empty objects in them, due to how we handle
            # the case above.  awkward syntax because pytorch 1.3 doesn't like
            # comparing tensors to dicts.
            if {} in [x for x in item.values() if isinstance(x, dict)]:
                raise ValueError(
                    'Cannot handle a dictionary with an empty dictionary inside.'
                )
            if () in [x for x in item.values() if isinstance(x, tuple)]:
                raise ValueError(
                    'Cannot handle a dictionary with an empty tuple inside.'
                )

            # we start with Dict[key,tensor]
            # we map it to d: Dict[key, List[Tensor]], where we have split each mapping
            d = {k: PipelineHelper.split(v, split_size, dim) for k, v in item.items()}
            # now we transpose it and return List[Dict[key, Tensor]]
            return [
                dict(zip(d.keys(), values))  # type: ignore
                for values in zip(*(d[k] for k in d.keys()))
            ]
        else:
            raise TypeError(f"Cannot split type {type(item)}")

    @staticmethod
    def join(items: List[Chunk], dim=0) -> Chunk:
        """
        Join chunks back together, the inverse of split.

        :param items:
            All the output chunks. Each chunk may be a tensor or a group of
            tensors.
        :param dim:
            The dimension to join along.
        """
        if len(items) == 0:
            raise IndexError("Cannot rejoin an empty list of chunks.")
        item0 = items[0]
        if isinstance(item0, torch.Tensor):
            # base case
            return torch.cat(items, dim=dim)  # type: ignore
        elif isinstance(item0, tuple):
            return tuple(
                PipelineHelper.join(x, dim=dim) for x in zip(*items)
            )  # type: ignore
        elif isinstance(item0, dict):
            keys = item0.keys()
            return {  # type: ignore
                k: PipelineHelper.join([c[k] for c in items], dim=dim)  # type: ignore
                for k in keys
            }
        else:
            raise TypeError(f'Cannot join list of type {type(item0)}')

    @staticmethod
    def chunk_to(chunk: Chunk, device: str) -> Chunk:
        """
        Move the chunk to the device.

        Handles chunks which are groups of tensors.
        """
        if isinstance(chunk, torch.Tensor):
            return chunk.to(device)  # type: ignore
        elif isinstance(chunk, tuple):
            return tuple(
                PipelineHelper.chunk_to(c, device) for c in chunk
            )  # type: ignore
        elif isinstance(chunk, dict):
            return {
                k: PipelineHelper.chunk_to(v, device) for k, v in chunk.items()
            }  # type: ignore
        else:
            raise TypeError('chunk_to only compatible with tensors, tuples or dicts.')

    @staticmethod
    def schedule_work_items(layers: torch.nn.ModuleList, chunks: List[Chunk]):
        """
        Iterate through chunks and layers that should be pipelined.

        Each iteration of this generator yields the following properties:

            - layer_nos: a list of indices of layers for you to forward through
            - chunk_idx: the index of the chunk we are manipulating. Use this
              if you need to update chunk representations.
            - next_device: where the chunk should be moved to AFTER the layer
              computation is done.
        """
        # We want to pipeline our computations so that each GPU is working on
        # chunks of the problem at the same of the time. The load of the will
        # look like this, assuming there are 5 chunks (A, B, C, D, E) and 4
        # GPUs. Each slot fill means that gpu is working on that chunk.
        #
        #         +-----------------+
        #         |       Time      |
        #         | 1 2 3 4 5 6 7 8 |
        # +-------+-----------------+
        # |  G  0 | A B C D E       |
        # |  P  1 |   A B C D E     |
        # |  U  2 |     A B C D E   |
        # |     3 |       A B C D E |
        # +-------+-----------------+
        #
        # Note that some GPUs will be idle much of the time. In reality, we
        # will use 2 * num_gpus as the number of chunks, to minimize idle
        # time.
        num_chunks = len(chunks)
        for l in layers:
            if not hasattr(l, '_mp_gpu'):
                raise RuntimeError(
                    'You must run PipelineHelper.make_parallel on the ModuleList '
                    'before you can use iterate_layers_chunks.'
                )

        # devices maps device_idx -> (device, [layer_idx, layer_idx, ...])
        # for example, if devices is 2 and there are 4 layers, we might have
        #   devices = {
        #     0: ('cuda:0', [0]),
        #     1: ('cuda:1', [1, 2, 3]),
        #   }
        # This means layers 0 is on cuda:0, but layers 1-3 are on cuda:1.
        devices = {
            device_idx: (dev, list(grp))
            for device_idx, (dev, grp) in enumerate(
                itertools.groupby(range(len(layers)), lambda x: layers[x]._mp_gpu)
            )
        }
        num_timesteps = len(devices) + num_chunks
        for timestep in range(num_timesteps):
            for chunk_idx in range(num_chunks):
                device_idx = timestep - chunk_idx
                if device_idx >= 0 and device_idx < len(devices):
                    dev, layers_nos = devices[device_idx]
                    next_device, _ = devices[(device_idx + 1) % len(devices)]
                    assert device_idx in devices
                    yield PipelineWorkItem(
                        chunk_idx=chunk_idx,
                        layer_nos=layers_nos,
                        next_device=next_device,
                    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher

from parlai.tasks.light_dialog.agents import DefaultTeacher as OrigLightTeacher
from parlai.tasks.light_genderation_bias.build import build
from parlai.utils.io import PathManager

from collections import deque
from copy import deepcopy
import csv
import json
import os
import random


# Filenames
NEW_DATA = 'new_data_dialogue_only.json'
GENDERED_LIST = 'gendered_list.tsv'
COUNTERFACTUALS = 'counterfactuals.json'


def _path(opt):
    build(opt)
    datapath = opt['datapath']
    return os.path.join(datapath, 'light_genderation_bias')


PUNCTUATION_LST = [
    (' .', '.'),
    (' !', '!'),
    (' ?', '?'),
    (' ,', ','),
    (" ' ", "'"),
    (" . . . ", "... "),
    (" ( ", " ("),
    (" ) ", ") "),
    (" ; ", "; "),
]


##############################################################
## UTILITY FUNCTIONS
##############################################################


def get_finegrained_count(text, gender_dct):
    """
    Count the number of female, male, and neutral gendered words in a string, given the
    gender dict.
    """
    text = text.lower()
    f_count = 0
    m_count = 0
    n_count = 0
    for line in text.split('\n'):
        words = line.split(' ')
        for word in words:
            if word in gender_dct:
                if gender_dct[word]['gender'] == 'F':
                    f_count += 1
                else:
                    m_count += 1
            else:
                n_count += 1

    return f_count, m_count, n_count


def format_text(text):
    text = text.lower()
    for punc in PUNCTUATION_LST:
        text = text.replace(punc[1], punc[0])

    return text


def unformat_text(text):
    for punc in PUNCTUATION_LST:
        text = text.replace(punc[0], punc[1])

    return text


def read_gender_tsv(path, remove_verbs=True):
    """
    Load TSV of gendered word lists and return a dict.
    """
    gender_dct = {}
    with PathManager.open(path) as tsvfile:
        reader = list(csv.reader(tsvfile, delimiter='\t'))
        title_lst = reader[0]
        title_dict = {}
        for idx, title in enumerate(title_lst):
            title_dict[idx] = title

        for i in range(1, len(reader)):
            row = reader[i]
            word = row[0].lower()
            gender_dct[word] = {}
            for j, category in enumerate(row[1:]):
                gender_dct[word][title_dict[j + 1]] = category

    if remove_verbs:
        return {k: v for k, v in gender_dct.items() if v['syncategory'] != 'verb'}

    return gender_dct


def flatten(episode, context_length, include_labels=True, delimiter='\n'):
    """
    Flatten the data into single example episodes.

    This is used to make conditional training easier and for a fair comparison of
    methods.
    """
    context = deque(maxlen=context_length if context_length > 0 else None)
    new_episode = []

    for ex in episode:
        context.append(ex.get('text', ''))
        # add context
        if len(context) > 1:
            ex.force_set('text', delimiter.join(context))
        # set episode_done to be True
        ex.force_set('episode_done', True)
        labels = ex.get('labels', ex.get('eval_labels', None))
        if labels is not None and include_labels:
            context.append(random.choice(labels))

        new_episode.append(ex)

    return new_episode


##############################################################
## TEACHERS FUNCTIONS
##############################################################
class LightGenderTeacher(FixedDialogTeacher):
    """
    ALL LIGHT Teacher: combines all gender bias mitigation methods described in.

    <https://arxiv.org/abs/1911.03842>.

    This teacher combines methods for gender bias mitigation in
    dialogue generation on the LIGHT dialogue task, including:
    - Adding more positive data (Pos Data):`--add-new-data`
    - Conditional training (Bias Ctrl): `--add-conditional`
    - Counterfactual Data Augmentation (CDA): `--add-counterfactual`

    For more information, please see our projects page:
    <https://parl.ai/projects/genderation_bias/>
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('LIGHT Gender Args')
        OrigLightTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        parser.add_argument(
            '--add-new-data',
            type='bool',
            default=True,
            help='Add the extra data collected',
        )
        parser.add_argument(
            '--add-conditional',
            type='bool',
            default=True,
            help='Append the conditional bucket to the end of the input or not',
        )
        parser.add_argument(
            '--add-counterfactual',
            type='bool',
            default=True,
            help='Load counterfactual data for training',
        )
        parser.add_argument(
            '--force-conditional',
            type=str,
            default='none',
            help='Add one bucket to everything, regardless of the true bin',
        )
        parser.add_argument(
            '--bucket-only',
            type=str,
            default='none',
            help='Only train/evaluate on examples from a specific bucket',
        )

        parser.set_defaults(
            light_use_objects=False,
            light_use_setting=False,
            light_use_affordances=False,
            light_use_persona_names=False,
            light_use_persona='none',
            light_use_action='none',
            light_use_emote='none',
            use_reply='none',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.gender_dict = read_gender_tsv(os.path.join(_path(opt), GENDERED_LIST))
        self.fixed_random = random.Random(42)

        self.add_conditional = opt['add_conditional']
        self.add_counterfactual = opt['add_counterfactual']
        self.add_new_data = opt['add_new_data']

        self.force_conditional = opt['force_conditional']
        self.bucket_only = opt['bucket_only']
        if self.force_conditional == 'none':
            self.force_conditional = None
        if self.bucket_only == 'none':
            self.bucket_only = None

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self.data = self._setup_data(opt)

        super().__init__(opt, shared)
        self.reset()

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def get_bucket(self, text):
        """
        Get the bin or "bucket" that a string falls into.

        The bucket is determined based on the presence of male or femael gendered words
        in the text.
        """
        text = format_text(text)
        # we're bucketing based on the number of gendered words
        f_count, m_count, n_count = get_finegrained_count(text, self.gender_dict)
        # get f bucket
        if f_count == 0:
            f_bucket = 'f0'
        else:
            f_bucket = 'f1'

        # get m bucket
        if m_count == 0:
            m_bucket = 'm0'
        else:
            m_bucket = 'm1'

        # combine bucket
        all_bucket = ' '.join([f_bucket, m_bucket])
        return all_bucket

    def _flip_str(self, txt_lines):
        new_lines = []
        lines = txt_lines.split('\n')
        for text in lines:
            f_text = format_text(text)
            f_text_lst = f_text.split(' ')
            new_words = []
            for word in f_text_lst:
                if word in self.swap_dct:
                    if word == 'her':
                        # silly heuristic: choose him/his 50% of the time
                        random_choice = random.choice([0, 1])
                        if random_choice:
                            new_word = 'his'
                        else:
                            new_word = 'him'
                    else:
                        new_word = self.swap_dct[word]['word']
                else:
                    new_word = word
                new_words.append(new_word)
            new_f_text = ' '.join(new_words)
            uf_text = unformat_text(new_f_text)
            new_lines.append(uf_text)

        return '\n'.join(new_lines)

    def _flip_ex(self, ex):
        """
        Return the counterfactual example for a given example (i.e. swap 'he' --> 'she')
        """
        new_ex = deepcopy(ex)
        text = ex['text']
        labels = 'labels' if 'labels' in ex else 'eval_labels'
        label = ex[labels][0]
        new_ex.force_set('text', self._flip_str(text))
        new_ex.force_set(labels, [self._flip_str(label)])
        return new_ex

    def _get_new_data(self, opt):
        """
        Load extra positive dialogue data IFF datatype==train.
        """
        dt = opt['datatype'].split(':')[0]
        if dt == 'train':
            with PathManager.open(os.path.join(_path(opt), NEW_DATA), 'r') as f:
                data = json.load(f)
            new_data = []
            for ep in data:
                new_ep = []
                for ex in ep:
                    ex['new_data'] = True
                    new_ep.append(Message(ex))
                new_data.append(new_ep)
            return new_data

        return []

    def _setup_data(self, opt):
        """
        Load original LIGHT dataset.
        """
        # Add new data?
        dt = opt['datatype'].split(':')[0]

        orig_episodes = OrigLightTeacher(opt).episodes
        if self.add_new_data:
            new_data = self._get_new_data(opt)
            total_data = orig_episodes + new_data
            self.fixed_random.shuffle(total_data)
            orig_episodes = total_data

        # Flatten this data
        flat_episodes = []
        for ep in orig_episodes:
            # flatten the episode into 1-example episodes with context
            flattened_ep = flatten(ep, -1, include_labels=True, delimiter='\n')
            flat_episodes += flattened_ep

        # Counterfactual?
        if self.add_counterfactual and dt != 'test':
            with PathManager.open(os.path.join(_path(opt), COUNTERFACTUALS), 'rb') as f:
                self.swap_dct = json.load(f)

            new_episodes = []
            for ex in flat_episodes:
                new_ex = self._flip_ex(ex)
                ex['counterfactual'] = False  # mark which episode is swapped
                new_ex['counterfactual'] = True
                # add both old and new examples
                new_episodes.append(ex)
                new_episodes.append(new_ex)

            flat_episodes = new_episodes

        # Conditional training?
        bucket_percentages = {}
        new_episodes = []
        for ex in flat_episodes:
            label_type = 'labels' if 'labels' in ex else 'eval_labels'
            label = ex[label_type][0]
            # get bucket for label
            bucket_key = self.get_bucket(label)
            # update the bucket percentages
            bucket_percentages.setdefault(bucket_key, 0)
            bucket_percentages[bucket_key] += 1
            # append this bucket to the text field
            if self.add_conditional:
                if self.force_conditional is None:
                    new_text = ex['text'] + '\n' + bucket_key
                else:
                    # force the model to see a specific bucket every time
                    # NOTE: we still track the original bucket that the
                    # text was supposed to fall into
                    new_text = ex['text'] + self.force_conditional
                ex.force_set('text', new_text)
            ex['bucket'] = bucket_key
            if self.bucket_only is None or self.bucket_only == bucket_key:
                new_episodes.append(ex)

        # Summarize the bucket distribution
        print('Distribution of bins:')
        total = sum(bucket_percentages.values())
        strs = []
        for k, v in bucket_percentages.items():
            pct = round((v / total) * 100, 2)
            strs.append(f'{k}: {pct}%')
        strs = sorted(strs)
        for string in strs:
            print(string)

        return new_episodes

    def get(self, episode_idx, entry_idx=0):
        return Message(self.data[episode_idx])

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared


class DefaultTeacher(LightGenderTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/COCO-IMG/train2017.zip',
        'train2017.zip',
        '69a8bb58ea5f8f99d24875f21416de2e9ded3178e903f1f7603e283b9e06d929',
    ),
    DownloadableFile(
        'http://parl.ai/downloads/COCO-IMG/val2017.zip',
        'val2017.zip',
        '4f7e2ccb2866ec5041993c9cf2a952bbed69647b115d0f74da7ce8f4bef82f05',
    ),
    DownloadableFile(
        'http://parl.ai/downloads/COCO-IMG/test2017.zip',
        'test2017.zip',
        'c7908c3c9f94ba2f3340ebbeec58c25db6be8774f18d68c2f15d0e369d95baba',
    ),
    DownloadableFile(
        'http://images.cocodataset.org/annotations/annotations_trainval2017.zip',
        'annotations_trainval2017.zip',
        '113a836d90195ee1f884e704da6304dfaaecff1f023f49b6ca93c4aaae470268',
    ),
    DownloadableFile(
        'http://images.cocodataset.org/annotations/image_info_test2017.zip',
        'image_info_test2017.zip',
        'e52f412dd7195ac8f98d782b44c6dd30ea10241e9f42521f67610fbe055a74f8',
    ),
]


def buildImage(opt):
    dpath = os.path.join(opt['datapath'], 'COCO-IMG-2017')
    version = '1'

    if not build_data.built(dpath, version_string=version):
        print('[building image data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the image data.
        for downloadable_file in RESOURCES[:3]:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)


def build(opt):
    dpath = os.path.join(opt['datapath'], 'COCO_2017_Caption')
    version = None

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(os.path.join(dpath, 'annotations'))
        # Make the subfolder into which files will be unzipped

        # download the data.
        for downloadable_file in RESOURCES[3:]:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.teachers import FbDeprecatedDialogTeacher, MultiTaskTeacher
from .build import build

import copy
import os


def _path(task, opt):
    # Build the data if it doesn't exist.
    build(opt)
    suffix = ''
    dt = opt['datatype'].split(':')[0]
    if dt == 'train':
        suffix = 'train'
    elif dt == 'test':
        suffix = 'test_2500ex'
    elif dt == 'valid':
        suffix = 'valid_2000ex'

    return os.path.join(
        opt['datapath'], 'CBT', 'CBTest', 'data', task + '_' + suffix + '.txt'
    )


class NETeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt['datafile'] = _path('cbtest_NE', opt)
        opt['cloze'] = True
        super().__init__(opt, shared)


class CNTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt['datafile'] = _path('cbtest_CN', opt)
        opt['cloze'] = True
        super().__init__(opt, shared)


class VTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt['datafile'] = _path('cbtest_V', opt)
        opt['cloze'] = True
        super().__init__(opt, shared)


class PTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt['datafile'] = _path('cbtest_P', opt)
        opt['cloze'] = True
        super().__init__(opt, shared)


# By default train on all tasks at once.
class DefaultTeacher(MultiTaskTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['task'] = 'cbt:NE,cbt:CN,cbt:V,cbt:P'
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
This helper script can be used alone with modelfile and task: the output will contain
the word statistics of the model outputs. One can also use the function defined here in
other places in order to get such statistic for any agent given the agent object (with
corr. dict) and a sequence.

Additionally provides function `get_word_stats` that can be used in
other parts of runtime code since it depends only on the agent object.
For example:

```python
from parlai.scripts.eval_wordstat import get_word_stats
reqs, cnt = get_word_stats(predictions.tolist(), self.dict)
```

## Examples

```shell
parlai eval_wordstat --model-file /path/to/model_file --task convai2:self --freq-bins 10,100,1000
```
"""

from parlai.core.params import ParlaiParser
from parlai.core.dict import DictionaryAgent
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.utils.misc import TimeLogger
from parlai.core.metrics import normalize_answer
from parlai.core.logs import TensorboardLogger
from collections import Counter
from parlai.core.script import ParlaiScript, register_script
from parlai.utils.io import PathManager

import copy
import numpy
import random


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, 'Compute statistics from model predictions')
    DictionaryAgent.add_cmdline_args(parser, partial_opt=None)
    # Get command line arguments
    parser.add_argument('-ne', '--num-examples', type=int, default=-1)
    parser.add_argument('-ltim', '--log-every-n-secs', type=float, default=2)
    parser.add_argument(
        '-ed',
        '--external-dict',
        type=str,
        default=None,
        help='External dictionary for stat computation',
    )
    parser.add_argument(
        '-fb',
        '--freq-bins',
        type=str,
        default='0,100,1000,10000',
        help='Bins boundaries for rare words stat',
    )
    parser.add_argument(
        '-dup',
        '--dump-predictions-path',
        type=str,
        default=None,
        help='Dump predictions into file',
    )
    parser.add_argument(
        '-cun',
        '--compute-unique',
        type='bool',
        default=True,
        help='Compute %% of unique responses from the model',
    )
    parser.set_defaults(datatype='valid')
    TensorboardLogger.add_cmdline_args(parser, partial_opt=None)
    return parser


def get_word_stats(text, agent_dict, bins=(0, 100, 1000, 100000)):
    """
    Function which takes text sequence and dict, returns word freq and length
    statistics.

    :param sequence: text sequence
    :param agent_dict: can be external dict or dict from the model
    :param bins: list with range boundaries
    :return: freqs dictionary, num words, avg word length, avg char length
    """
    pred_list = agent_dict.tokenize(text)
    pred_freq = [agent_dict.freq[word] for word in pred_list]
    freqs = {i: 0 for i in bins}
    for f in pred_freq:
        for b in bins:
            if f <= b:
                freqs[b] += 1
                break

    wlength = len(pred_list)
    clength = len(text)  # including spaces
    return freqs, len(pred_freq), wlength, clength


def eval_wordstat(opt):
    """
    Evaluates a model.

    :param opt: tells the evaluation function how to run
    """
    random.seed(42)

    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    world = create_task(opt, agent)
    agent.opt.log()

    if opt.get('external_dict'):
        print('[ Using external dictionary from: {} ]'.format(opt['external_dict']))
        dict_opt = copy.deepcopy(opt)
        dict_opt['dict_file'] = opt['external_dict']
        dictionary = DictionaryAgent(dict_opt)
    else:
        print('[ Using model bundled dictionary ]')
        dictionary = agent.dict

    batch_size = opt['batchsize']

    log_every_n_secs = opt.get('log_every_n_secs', -1)
    if log_every_n_secs <= 0:
        log_every_n_secs = float('inf')
    log_time = TimeLogger()

    cnt = 0
    max_cnt = opt['num_examples'] if opt['num_examples'] > 0 else float('inf')
    word_statistics = {
        'mean_wlength': [],
        'mean_clength': [],
        'freqs_cnt': Counter(),
        'word_cnt': 0,
        'pred_list': [],
        'pure_pred_list': [],
        'context_list': [],
        'unique_words': set(),
    }
    bins = [int(i) for i in opt['freq_bins'].split(',')]

    def process_prediction(prediction, word_statistics):
        normalized = normalize_answer(prediction)
        word_statistics['pred_list'].append(normalized)
        freqs, _cnt, wlength, clength = get_word_stats(
            prediction, dictionary, bins=bins
        )
        word_statistics['word_cnt'] += _cnt
        word_statistics['mean_wlength'].append(wlength)
        word_statistics['mean_clength'].append(clength)
        word_statistics['freqs_cnt'] += Counter(freqs)
        word_statistics['unique_words'] |= set(normalized.split(" "))
        return word_statistics

    while not world.epoch_done():
        world.parley()
        if batch_size == 1:
            cnt += 1
            prediction = world.acts[-1]['text']
            word_statistics['context_list'].append(world.acts[0]['text'])
            word_statistics['pure_pred_list'].append(prediction)
            word_statistics = process_prediction(prediction, word_statistics)
        else:
            for w in world.worlds:
                try:
                    if 'text' not in w.acts[-1]:
                        continue
                    prediction = w.acts[-1]['text']
                    word_statistics['context_list'].append(w.acts[0]['text'])
                    word_statistics['pure_pred_list'].append(prediction)
                except IndexError:
                    continue
                cnt += 1
                word_statistics = process_prediction(prediction, word_statistics)

        if log_time.time() > log_every_n_secs:
            report = world.report()
            text, report = log_time.log(
                report['exs'], min(max_cnt, world.num_examples()), report
            )
            print(text)
            stat_str = 'total_words: {}, '.format(word_statistics['word_cnt'])
            stat_str += ', '.join(
                [
                    '<{}:{} ({:.{prec}f}%)'.format(
                        b,
                        word_statistics['freqs_cnt'].get(b, 0),
                        (
                            word_statistics['freqs_cnt'].get(b, 0)
                            / word_statistics['word_cnt']
                        )
                        * 100,
                        prec=2,
                    )
                    for b in bins
                ]
            )
            print(
                "Word statistics: {}, avg_word_length: {:.{prec}f}, "
                "avg_char_length: {:.{prec}f}".format(
                    stat_str,
                    numpy.array(word_statistics['mean_wlength']).mean(),
                    numpy.array(word_statistics['mean_clength']).mean(),
                    prec=2,
                )
            )
        if cnt >= max_cnt:
            break
    if world.epoch_done():
        print("EPOCH DONE")

    if opt['compute_unique'] is True:
        unique_list = []
        cntr = Counter(word_statistics['pred_list'])
        for k, v in cntr.items():
            if v == 1:
                unique_list.append(k)
        print(
            "Unique responses: {:.{prec}f}%".format(
                len(unique_list) / len(word_statistics['pred_list']) * 100, prec=2
            )
        )
    print("Total unique tokens:", len(word_statistics['unique_words']))

    if opt['dump_predictions_path'] is not None:
        with PathManager.open(opt['dump_predictions_path'], 'w') as f:
            f.writelines(
                [
                    'CONTEXT: {}\nPREDICTION:{}\n\n'.format(c, p)
                    for c, p in zip(
                        word_statistics['context_list'],
                        word_statistics['pure_pred_list'],
                    )
                ]
            )
        if opt['compute_unique'] is True:
            with PathManager.open(opt['dump_predictions_path'] + '_unique', 'w') as f:
                f.writelines(['{}\n'.format(i) for i in unique_list])

    stat_str = 'total_words: {}, '.format(word_statistics['word_cnt'])
    stat_str += ', '.join(
        [
            '<{}:{} ({:.{prec}f}%)'.format(
                b,
                word_statistics['freqs_cnt'].get(b, 0),
                (word_statistics['freqs_cnt'].get(b, 0) / word_statistics['word_cnt'])
                * 100,
                prec=2,
            )
            for b in bins
        ]
    )
    print(
        "Word statistics: {}, avg_word_length: {:.{prec}f}, "
        "avg_char_length: {:.{prec}f}".format(
            stat_str,
            numpy.array(word_statistics['mean_wlength']).mean(),
            numpy.array(word_statistics['mean_clength']).mean(),
            prec=2,
        )
    )

    report = world.report()
    print(report)
    return report


@register_script('eval_wordstat', hidden=True)
class EvalWordStat(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return eval_wordstat(self.opt)


if __name__ == '__main__':
    EvalWordStat.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from typing import Optional

import gzip
import os

from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data


class ParseInsuranceQA(object):
    version: Optional[str] = None
    label2answer_fname: Optional[str] = None

    @classmethod
    def read_gz(cls, filename):
        f = gzip.open(filename, 'rb')
        return [x.decode('utf-8') for x in f.readlines()]

    @classmethod
    def readlines(cls, path):
        if path.endswith(".gz"):
            lines = cls.read_gz(path)
        else:
            lines = open(path).readlines()
        return lines

    @classmethod
    def wids2sent(cls, wids, d_vocab):
        return " ".join([d_vocab[w] for w in wids])

    @classmethod
    def read_vocab(cls, vocab_path):
        d_vocab = {}
        with PathManager.open(vocab_path, "r") as f:
            for line in f:
                fields = line.rstrip('\n').split("\t")
                if len(fields) != 2:
                    raise ValueError(
                        "vocab file (%s) corrupted. Line (%s)"
                        % (repr(line), vocab_path)
                    )
                else:
                    wid, word = fields
                    d_vocab[wid] = word
        return d_vocab

    @classmethod
    def read_label2answer(cls, label2answer_path_gz, d_vocab):
        lines = cls.readlines(label2answer_path_gz)

        d_label_answer = {}
        for line in lines:
            fields = line.rstrip("\n").split("\t")
            if len(fields) != 2:
                raise ValueError(
                    "label2answer file (%s) corrupted. Line (%s)"
                    % (repr(line), label2answer_path_gz)
                )
            else:
                aid, s_wids = fields
                sent = cls.wids2sent(s_wids.split(), d_vocab)
                d_label_answer[aid] = sent
        return d_label_answer

    @classmethod
    def create_fb_format(cls, out_path, dtype, inpath, d_vocab, d_label_answer):
        pass

    @classmethod
    def write_data_files(cls, dpext, out_path, d_vocab, d_label_answer):
        pass

    @classmethod
    def build(cls, dpath):
        print("building version: %s" % cls.version)

        # the root of dataset
        dpext = os.path.join(dpath, 'insuranceQA-master/%s' % cls.version)

        # read vocab file
        vocab_path = os.path.join(dpext, "vocabulary")
        d_vocab = cls.read_vocab(vocab_path)

        # read label2answer file
        label2answer_path_gz = os.path.join(dpext, cls.label2answer_fname)
        d_label_answer = cls.read_label2answer(label2answer_path_gz, d_vocab)

        # Create out path
        out_path = os.path.join(dpath, cls.version)
        build_data.make_dir(out_path)

        # Parse and write data files
        cls.write_data_files(dpext, out_path, d_vocab, d_label_answer)


class ParseInsuranceQAV1(ParseInsuranceQA):
    version = "V1"
    label2answer_fname = "answers.label.token_idx"

    @classmethod
    def write_data_files(cls, dpext, out_path, d_vocab, d_label_answer):
        data_fnames = [
            ("train", "question.train.token_idx.label"),
            ("valid", "question.dev.label.token_idx.pool"),
            ("test", "question.test1.label.token_idx.pool"),
            # ("test2", "question.test2.label.token_idx.pool")
        ]
        for dtype, data_fname in data_fnames:
            data_path = os.path.join(dpext, data_fname)
            cls.create_fb_format(out_path, dtype, data_path, d_vocab, d_label_answer)

    @classmethod
    def create_fb_format(cls, out_path, dtype, inpath, d_vocab, d_label_answer):
        print('building fbformat:' + dtype)
        fout = open(os.path.join(out_path, dtype + '.txt'), 'w')
        lines = open(inpath).readlines()

        for line in lines:
            fields = line.rstrip("\n").split("\t")
            if dtype == "train":
                assert len(fields) == 2, "data file (%s) corrupted." % inpath
                s_q_wids, s_good_aids = fields

                q = cls.wids2sent(s_q_wids.split(), d_vocab)
                good_ans = [d_label_answer[aid_] for aid_ in s_good_aids.split()]
                # save good answers (train only)
                s = '1 ' + q + '\t' + "|".join(good_ans)
                fout.write(s + '\n')
            else:
                assert len(fields) == 3, "data file (%s) corrupted." % inpath
                s_good_aids, s_q_wids, s_bad_aids = fields

                q = cls.wids2sent(s_q_wids.split(), d_vocab)
                good_ans = [d_label_answer[aid_] for aid_ in s_good_aids.split()]
                bad_ans = [d_label_answer[aid_] for aid_ in s_bad_aids.split()]
                # save good answers and candidates
                s = (
                    '1 '
                    + q
                    + '\t'
                    + "|".join(good_ans)
                    + '\t\t'
                    + "|".join(good_ans + bad_ans)
                )
                fout.write(s + '\n')
        fout.close()


class ParseInsuranceQAV2(ParseInsuranceQA):
    version = "V2"
    label2answer_fname = "InsuranceQA.label2answer.token.encoded.gz"

    @classmethod
    def write_data_files(cls, dpext, out_path, d_vocab, d_label_answer):
        data_fnames_tmpl = [
            (
                "train.%s",
                "InsuranceQA.question.anslabel.token.%s.pool.solr.train.encoded.gz",
            ),  # noqa: E501
            (
                "valid.%s",
                "InsuranceQA.question.anslabel.token.%s.pool.solr.valid.encoded.gz",
            ),  # noqa: E501
            (
                "test.%s",
                "InsuranceQA.question.anslabel.token.%s.pool.solr.test.encoded.gz",
            ),  # noqa: E501
        ]
        for n_cands in [100, 500, 1000, 1500]:
            for dtype_tmp, data_fname_tmp in data_fnames_tmpl:
                dtype = dtype_tmp % n_cands
                data_fname = data_fname_tmp % n_cands
                data_path = os.path.join(dpext, data_fname)
                cls.create_fb_format(
                    out_path, dtype, data_path, d_vocab, d_label_answer
                )

    @classmethod
    def create_fb_format(cls, out_path, dtype, inpath, d_vocab, d_label_answer):
        print('building fbformat:' + dtype)
        fout = open(os.path.join(out_path, dtype + '.txt'), 'w')
        lines = cls.readlines(inpath)

        for line in lines:
            fields = line.rstrip("\n").split("\t")
            if len(fields) != 4:
                raise ValueError(
                    "data file (%s) corrupted. Line (%s)" % (repr(line), inpath)
                )
            else:
                _, s_q_wids, s_good_aids, s_bad_aids = fields
                q = cls.wids2sent(s_q_wids.split(), d_vocab)
                good_ans = [d_label_answer[aid_] for aid_ in s_good_aids.split()]
                bad_ans = [d_label_answer[aid_] for aid_ in s_bad_aids.split()]
                # save
                s = (
                    '1 '
                    + q
                    + '\t'
                    + "|".join(good_ans)
                    + '\t\t'
                    + "|".join(good_ans + bad_ans)
                )
                fout.write(s + '\n')
        fout.close()


RESOURCES = [
    DownloadableFile(
        'https://github.com/shuzi/insuranceQA/archive/master.zip',
        'insuranceqa.zip',
        '53e1c4a68734c6a0955dcba50d5a2a9926004d4cd4cda2e988cc7b990a250fbf',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'InsuranceQA')
    version = '1'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        ParseInsuranceQAV1.build(dpath)
        ParseInsuranceQAV2.build(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data
import os
import json

VERSION = '1'
TRAIN_FILENAME = 'hotpot_train_v{}.1.json'.format(VERSION)
DEV_DISTRACTOR_FILENAME = 'hotpot_dev_distractor_v{}.json'.format(VERSION)
DEV_FULLWIKI_FILENAME = 'hotpot_dev_fullwiki_v{}.json'.format(VERSION)

RESOURCES = [
    DownloadableFile(
        'http://curtis.ml.cmu.edu/datasets/hotpot/hotpot_train_v1.1.json',
        'hotpot_train_v1.1.json',
        '26650cf50234ef5fb2e664ed70bbecdfd87815e6bffc257e068efea5cf7cd316',
        zipped=False,
    ),
    DownloadableFile(
        'http://curtis.ml.cmu.edu/datasets/hotpot/hotpot_dev_distractor_v1.json',
        'hotpot_dev_distractor_v1.json',
        '4e9ecb5c8d3b719f624d66b60f8d56bf227f03914f5f0753d6fa1b359d7104ea',
        zipped=False,
    ),
    DownloadableFile(
        'http://curtis.ml.cmu.edu/datasets/hotpot/hotpot_dev_fullwiki_v1.json',
        'hotpot_dev_fullwiki_v1.json',
        '2f1f3e594a3066a3084cc57950ca2713c24712adaad03af6ccce18d1846d5618',
        zipped=False,
    ),
]

OUTPUT_FORMAT = 'text:{context_question}\t' 'labels:{answer}'


def _handle_data_point(data_point):
    output = []
    context_question_txt = ""
    for [title, sentences_list] in data_point['context']:
        sentences = '\\n'.join(sentences_list)
        context_question_txt += '{}\\n{}\\n\\n'.format(title, sentences)

    context_question_txt += data_point['question']

    output = OUTPUT_FORMAT.format(
        context_question=context_question_txt, answer=data_point['answer']
    )
    output += '\t\tepisode_done:True\n'
    return output


def make_parlai_format(outpath, dtype, data):
    print('building parlai:' + dtype)
    with PathManager.open(os.path.join(outpath, dtype + '.txt'), 'w') as fout:
        for data_point in data:
            fout.write(_handle_data_point(data_point))


def build(opt):
    dpath = os.path.join(opt['datapath'], 'HotpotQA')

    if not build_data.built(dpath, version_string=VERSION):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        with PathManager.open(os.path.join(dpath, TRAIN_FILENAME)) as f:
            data = json.load(f)
            make_parlai_format(dpath, 'train', data)

        with PathManager.open(os.path.join(dpath, DEV_DISTRACTOR_FILENAME)) as f:
            data = json.load(f)
            make_parlai_format(dpath, 'valid_distractor', data)

        with PathManager.open(os.path.join(dpath, DEV_FULLWIKI_FILENAME)) as f:
            data = json.load(f)
            make_parlai_format(dpath, 'valid_fullwiki', data)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=VERSION)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher
from .build import build

import xml.etree.ElementTree as ET
import os
import copy


COPA = "COPA"
COPA_RESOURCES_FOLDER_NAME = 'COPA-resources'
COPA_DATASETS_FOLDER_NAME = 'datasets'
COPA_DATASET_PREFIX = 'copa-'
COPA_CAUSE_SUFFIX = "What was the CAUSE for this?"
COPA_RESULT_SUFFIX = "What happened as a RESULT?"


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train' or dt == 'valid':
        suffix = 'dev'
    elif dt == 'test':
        suffix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(
        opt['datapath'],
        COPA,
        COPA_RESOURCES_FOLDER_NAME,
        COPA_DATASETS_FOLDER_NAME,
        COPA_DATASET_PREFIX + suffix + '.xml',
    )

    return data_path


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'COPA'

        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading: ' + path)

        tree = ET.parse(path)
        root = tree.getroot()

        for child in root:
            asks_for = child.attrib['asks-for']
            answer = child.attrib['most-plausible-alternative']

            premise = child[0].text
            alternative_one = child[1].text
            alternative_two = child[2].text

            if asks_for == "cause":
                premise += " " + COPA_CAUSE_SUFFIX
            else:
                premise += " " + COPA_RESULT_SUFFIX

            if answer == "1":
                answer = [alternative_one]
            else:
                answer = [alternative_two]
            answer_options = [alternative_one, alternative_two]

            yield (premise, answer, None, answer_options), True



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager
from .build import build
import os
import json


def _path(opt):
    # build the data if it does not exist
    build(opt)

    # set up path to data (specific to each dataset)
    jsons_path = os.path.join(opt['datapath'], 'multiwoz_v20', 'MULTIWOZ2 2')
    conversations_path = os.path.join(jsons_path, 'data.json')
    return conversations_path, jsons_path


class MultiWozTeacher(FixedDialogTeacher):
    """
    MultiWOZ 2.0 Teacher.

    This dataset contains more than just dialogue. It also contains:
    data.json also contains the following information:
    1. information about the goal of the conversation (what the person is looking for)
    2. topic of the conversation
    3. log of the conversation (the conversation itself + metadata about each utterance)
          Metadata: any taxi, police, restaurant, hospital, hotel, attraction, or train info mentioned
    Information about each metadata category is also contained in its own json file.
    1. restaurant.json: Restaurants + their attributes in the Cambridge UK area (address, pricerange, food...)
    2. attraction.json: Attractions + their attributes in the Cambridge UK area (location, address, entrance fee...)
    3. hotel_db.json: Hotels + their attributes in the Cambridge UK area  (address, price, name, stars...)
    4. train_db.json: Trains + their attributes in the Cambridge UK area (destination, price, departure...)
    5. hospital_db.json: data about the Cambridge hospital's departments (department, phone, id)
    6. police_db.json: Name address and phone number of Cambridge police station
    7. taxi_db.json: Taxi information (taxi_colors, taxi_types, taxi_phone)
    More information about the jsons can be found in readme.json
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        opt['datafile'], jsons_path = _path(opt)
        self._setup_data(opt['datafile'], jsons_path)
        self.id = 'multiwoz_v20'
        self.reset()

    def _setup_data(self, data_path, jsons_path):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.messages = json.load(data_file)

        test_path = os.path.join(jsons_path, 'testListFile.json')
        valid_path = os.path.join(jsons_path, 'valListFile.json')
        if self.datatype.startswith('test'):
            with PathManager.open(test_path) as f:
                test_data = {line.strip(): self.messages[line.strip()] for line in f}
                self.messages = test_data
        elif self.datatype.startswith('valid'):
            with PathManager.open(valid_path) as f:
                valid_data = {line.strip(): self.messages[line.strip()] for line in f}
                self.messages = valid_data
        else:
            with PathManager.open(test_path) as f:
                for line in f:
                    if line.strip() in self.messages:
                        del self.messages[line.strip()]
            with PathManager.open(valid_path) as f:
                for line in f:
                    if line.strip() in self.messages:
                        del self.messages[line.strip()]
        self.messages = list(self.messages.values())

    def num_examples(self):
        examples = 0
        for data in self.messages:
            examples += len(data['log']) // 2
        return examples

    def num_episodes(self):
        return len(self.messages)

    def get(self, episode_idx, entry_idx=0):
        log_idx = entry_idx * 2
        entry = self.messages[episode_idx]['log'][log_idx]['text']
        episode_done = log_idx == len(self.messages[episode_idx]['log']) - 2
        action = {
            'id': self.id,
            'text': entry,
            'episode_done': episode_done,
            'labels': [self.messages[episode_idx]['log'][log_idx + 1]['text']],
        }
        return action


class DefaultTeacher(MultiWozTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This downloads a pretrained language model BART (Lewis et al.

https://arxiv.org/abs/1910.13461). It requires you to run a conversion script to map
model weights (conversion script not provided publicly).
"""

from parlai.agents.bart.convert_fairseq_to_parlai import ConversionScript
import parlai.core.build_data as build_data
import os

CONVERSION_ARGS = {
    'add_prefix_space': False,
    'activation': 'gelu',
    'tokenizer': 'gpt2',
    'delimiter': '\n',
    'retain_bos_emb': True,
    'model': 'bart',
    'fp16': True,
    'history_add_global_end_token': None,
}

BART_ARGS = {
    'embedding_size': 1024,
    'ffn_size': 4096,
    'dropout': 0.1,
    'attention_dropout': 0.1,
    'n_heads': 16,
    'n_positions': 1024,
    'variant': 'bart',
    'activation': 'gelu',
    'n_encoder_layers': 12,
    'n_decoder_layers': 12,
    'force_fp16_tokens': True,
    'fp16': True,
    'dict_tokenizer': 'gpt2',
    'embeddings_scale': False,
    'history_add_global_end_token': None,
    'learn_positional_embeddings': True,
}


def download(datapath, version='v1.0'):
    dpath = os.path.join(datapath, 'models', 'bart')

    if not build_data.built(dpath, version):
        print('[downloading BART models: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        model_name = 'bart.large'
        url = f'http://dl.fbaipublicfiles.com/fairseq/models/{model_name}.tar.gz'
        build_data.download(url, dpath, f'{model_name}.tar.gz')
        build_data.untar(dpath, f'{model_name}.tar.gz')
        args = CONVERSION_ARGS.copy()
        args['input'] = [os.path.join(dpath, model_name, 'model.pt')]
        args['output'] = os.path.join(dpath, model_name.replace('.', '_'), 'model')
        ConversionScript.main(**args)

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import json
import os
import re
import numpy as np
from collections import Counter

from parlai.core.agents import Agent
from parlai.utils.io import PathManager
from collections import defaultdict
from parlai.core.teachers import FixedDialogTeacher
from parlai.core.image_featurizers import ImageLoader
from .build import build
from parlai.tasks.coco_caption.build_2014 import buildImage as buildImage_2014
from parlai.tasks.coco_caption.build_2015 import buildImage as buildImage_2015


def _path(opt):
    build(opt)
    buildImage_2014(opt)
    buildImage_2015(opt)
    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        ques_suffix = 'MultipleChoice_mscoco_train2014'
        annotation_suffix = 'mscoco_train2014'
        img_suffix = os.path.join('train2014', 'COCO_train2014_')
        img_version = '2014'
    elif dt == 'valid':
        ques_suffix = 'MultipleChoice_mscoco_val2014'
        annotation_suffix = 'mscoco_val2014'
        img_suffix = os.path.join('val2014', 'COCO_val2014_')
        img_version = '2014'
    elif dt == 'test':
        ques_suffix = 'MultipleChoice_mscoco_test2015'
        annotation_suffix = 'None'
        img_suffix = os.path.join('test2015', 'COCO_test2015_')
        img_version = '2015'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], 'VQA-v1', ques_suffix + '_questions.json')

    annotation_path = os.path.join(
        opt['datapath'], 'VQA-v1', annotation_suffix + '_annotations.json'
    )

    image_path = os.path.join(
        opt['datapath'], 'COCO-IMG-{}'.format(img_version), img_suffix
    )

    return data_path, annotation_path, image_path


def escape(s):
    """
    Replace potential special characters with escaped version.

    For example, newline => \\n and tab => \\t
    """
    return s.replace('\n', '\\n').replace('\t', '\\t').replace('\r', '\\r')


def unescape(s):
    """
    Revert escaped characters back to their special version.

    For example, \\n => newline and \\t => tab
    """
    return s.replace('\\n', '\n').replace('\\t', '\t').replace('\\r', '\r')


class VqaDictionaryAgent(Agent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        dictionary = parser.add_argument_group('Dictionary Arguments')
        dictionary.add_argument(
            '--dict-file',
            help='if set, the dictionary will automatically save to this path'
            + ' during shutdown',
        )
        dictionary.add_argument(
            '--dict-initpath',
            help='path to a saved dictionary to load tokens / counts from to '
            + 'seed the dictionary with initial tokens and/or frequencies',
        )
        dictionary.add_argument(
            '--dict-maxexs',
            default=300000,
            type=int,
            help='max number of examples to build dict on',
        )
        dictionary.add_argument('-smp', '--samplingans', type='bool', default=True)
        dictionary.add_argument('--nans', type=int, default=2000)
        dictionary.add_argument('--maxlength', type=int, default=16)
        dictionary.add_argument('--minwcount', type=int, default=0)
        dictionary.add_argument('--nlp', default='mcb')
        return parser

    def __init__(self, opt, shared=None):
        super(VqaDictionaryAgent, self).__init__(opt)
        self.id = 'VqaDictionary'
        self.null_token = '__NULL__'
        self.unk_token = '__UNK__'

        if shared:
            self.freq = shared.get('freq', {})
            self.tok2ind = shared.get('tok2ind', {})
            self.ind2tok = shared.get('ind2tok', {})
            self.ans2ind = shared.get('ans2ind', {})
            self.ind2ans = shared.get('ind2ans', {})
        else:
            self.freq = defaultdict(int)
            self.ansfreq = defaultdict(int)
            self.ans2ques = defaultdict(list)
            self.tok2ind = {}
            self.ind2tok = {}
            self.ans2ind = {}
            self.ind2ans = {}

            if self.null_token:
                self.tok2ind[self.null_token] = 0
                self.ind2tok[0] = self.null_token

            if self.unk_token:
                # set special unknown word token
                index = len(self.tok2ind)
                self.tok2ind[self.unk_token] = index
                self.ind2tok[index] = self.unk_token

        if opt.get('dict_file') and PathManager.exists(opt['dict_file']):
            # load pre-existing dictionary
            self.load(opt['dict_file'])

        if not shared:

            if self.null_token:
                # fix count for null token to one billion and two
                self.freq[self.null_token] = 1000000002

            if self.unk_token:
                # fix count for unknown token to one billion
                self.freq[self.unk_token] = 1000000000

            if opt.get('dict_file'):
                self.save_path = opt['dict_file']

    def __len__(self):
        return len(self.tok2ind)

    def add_to_ques_dict(self, tokens):
        """
        Builds dictionary from the list of provided tokens.

        Only adds words contained in self.embedding_words, if not None.
        """
        for token in tokens:
            self.freq[token] += 1
            if token not in self.tok2ind:
                index = len(self.tok2ind)
                self.tok2ind[token] = index
                self.ind2tok[index] = token

    def add_to_ans_dict(self, token):
        """
        Builds dictionary from the list of provided tokens.

        Only adds words contained in self.embedding_words, if not None.
        """
        self.ansfreq[token] += 1
        if token not in self.ans2ind:
            index = len(self.ans2ind)
            self.ans2ind[token] = index
            self.ind2ans[index] = token

    def tokenize_mcb(self, s):
        t_str = s.lower()
        for i in [
            r'\?',
            r'\!',
            r'\'',
            r'\"',
            r'\$',
            r'\:',
            r'\@',
            r'\(',
            r'\)',
            r'\,',
            r'\.',
            r'\;',
        ]:
            t_str = re.sub(i, '', t_str)
        for i in [r'\-', r'\/']:
            t_str = re.sub(i, ' ', t_str)
        q_list = re.sub(r'\?', '', t_str.lower()).split(' ')
        q_list = list(filter(lambda x: len(x) > 0, q_list))
        return q_list

    def split_tokenize(self, s):
        return (
            s.lower()
            .replace('.', ' . ')
            .replace('. . .', '...')
            .replace(',', ' , ')
            .replace(';', ' ; ')
            .replace(':', ' : ')
            .replace('!', ' ! ')
            .replace('?', ' ? ')
            .split()
        )

    def act(self):
        """
        Add any words passed in the 'text' field of the observation to this dictionary.
        """
        mc_label = self.observation.get('mc_label', self.observation.get('labels', []))
        for text in mc_label:
            self.ansfreq[text] += 1
            self.ans2ques[text].append(self.tokenize_mcb(self.observation.get('text')))
        return {'id': 'Dictionary'}

    def encode_question(self, examples, training):
        minwcount = self.opt.get('minwcount', 0)
        maxlength = self.opt.get('maxlength', 16)
        for ex in examples:
            words = self.tokenize_mcb(ex['text'])
            if training:
                words_unk = [
                    w if self.freq.get(w, 0) > minwcount else self.unk_token
                    for w in words
                ]
            else:
                words_unk = [w if w in self.tok2ind else self.unk_token for w in words]
            ex['question_wids'] = [self.tok2ind[self.null_token]] * maxlength
            for k, w in enumerate(words_unk):
                if k < maxlength:
                    ex['question_wids'][k] = self.tok2ind[w]
        return examples

    def encode_answer(self, examples):
        for ex in examples:
            if self.opt.get('samplingans', True):
                labels = ex.get('labels', ex.get('eval_labels'))
                ans_count = Counter(labels).most_common()
                valid_ans = []
                valid_count = []
                for ans in ans_count:
                    if ans[0] in self.ans2ind:
                        valid_ans.append(self.ans2ind[ans[0]])
                        valid_count.append(ans[1])
                if not valid_ans:
                    ex['answer_aid'] = 0
                else:
                    probs = valid_count / np.sum(valid_count)
                    ex['answer_aid'] = int(np.random.choice(valid_ans, p=probs))
            else:
                ex['answer_aid'] = self.ans2ind[ex['mc_label'][0]]
        return examples

    def decode_answer(self, examples):
        txt_answers = []
        for ex in examples:
            txt_answers.append(self.ind2ans[ex])
            # print("Predicted output ex:", i, ex)
        return txt_answers

    def load(self, filename):
        """
        Load pre-existing dictionary in 'token[<TAB>count]' format.

        Initialize counts from other dictionary, or 0 if they aren't included.
        """
        print('Dictionary: loading dictionary from {}'.format(filename))
        with PathManager.open(filename) as read:
            for line in read:
                split = line.strip().split('\t')
                token = unescape(split[0])
                cnt = int(split[1]) if len(split) > 1 else 0
                self.freq[token] = cnt
                if token not in self.tok2ind:
                    index = len(self.tok2ind)
                    self.tok2ind[token] = index
                    self.ind2tok[index] = token
        print('[ num ques words =  %d ]' % len(self.ind2tok))

        with PathManager.open(filename[:-5] + "_ans.dict") as read:
            for line in read:
                split = line.strip().split('\t')
                token = unescape(split[0])
                cnt = int(split[1]) if len(split) > 1 else 0
                self.ansfreq[token] = cnt
                if token not in self.ans2ind:
                    index = len(self.ans2ind)
                    self.ans2ind[token] = index
                    self.ind2ans[index] = token

        print('[ num ans words =  %d ]' % len(self.ind2ans))

    def save(self, filename=None, append=False, sort=True):
        """
        Save dictionary to file. Format is 'token<TAB>count' for every token in the
        dictionary, sorted by count with the most frequent words first.

        If ``append`` (default ``False``) is set to ``True``, appends instead
        of overwriting.

        If ``sort`` (default ``True``), then first sort the dictionary before
        saving.
        """
        cw = sorted([(count, w) for w, count in self.ansfreq.items()], reverse=True)
        final_exs = cw[: self.opt.get('nans', 2000)]
        final_list = dict([(w, c) for c, w in final_exs])
        self.ansfreq = defaultdict(int)
        for ans, ques in self.ans2ques.items():
            if ans in final_list:
                for que in ques:
                    self.add_to_ques_dict(que)
                self.add_to_ans_dict(ans)

        filename = self.opt['dict_file'] if filename is None else filename
        print('Dictionary: saving dictionary to {}'.format(filename))
        # if sort:
        #     self.sort()

        with PathManager.open(filename, 'a' if append else 'w') as write:
            for i in range(len(self.ind2tok)):
                tok = self.ind2tok[i]
                cnt = self.freq[tok]
                write.write('{tok}\t{cnt}\n'.format(tok=escape(tok), cnt=cnt))

        with PathManager.open(
            filename[:-5] + "_ans.dict", 'a' if append else 'w'
        ) as write:
            for i in range(len(self.ind2ans)):
                tok = self.ind2ans[i]
                cnt = self.ansfreq[tok]
                write.write('{tok}\t{cnt}\n'.format(tok=escape(tok), cnt=cnt))

    def shutdown(self):
        """
        Save on shutdown if ``save_path`` is set.
        """
        if hasattr(self, 'save_path'):
            self.save(self.save_path)


class OeTeacher(FixedDialogTeacher):
    """
    VQA Open-Ended teacher, which loads the json vqa data and implements its own `act`
    method for interacting with student agent.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        data_path, annotation_path, self.image_path = _path(opt)
        self.datafile = data_path
        self.image_mode = opt.get('image_mode', 'no_image_model')

        if shared and 'ques' in shared:
            self.ques = shared['ques']
            if 'annotation' in shared:
                self.annotation = shared['annotation']
            self.image_loader = shared['image_loader']
        else:
            self._setup_data(data_path, annotation_path)
            self.image_loader = ImageLoader(opt)
        self.reset()

    def reset(self):
        super().reset()
        self.example = None
        self.imageEpochDone = False

    def num_examples(self):
        """
        Number of examples in VQA-v1.
        """
        return len(self.ques['questions'])

    def num_episodes(self):
        # same as number of examples since all episodes are of length one
        return self.num_examples()

    def submit_load_request(self, image_id):
        img_path = self.image_path + '%012d.jpg' % (image_id)
        self.data_loader.request_load(
            self.receive_data, self.image_loader.load, (img_path,)
        )

    def get(self, episode_idx, entry_idx=0):
        # queue up the next one
        qa = self.ques['questions'][episode_idx]
        question = qa['question']

        action = {'text': question, 'image_id': qa['image_id'], 'episode_done': True}

        if not self.datatype.startswith('test'):
            anno = self.annotation['annotations'][episode_idx]
            action['labels'] = [ans['answer'] for ans in anno['answers']]

        return action

    def next_example(self):
        """
        Returns the next example from this dataset after starting to queue up the next
        example.
        """
        ready = None
        # pull up the currently queued example
        if self.example is not None:
            if self.image_mode != 'no_image_model' and 'image_id' in self.example:
                # move the image we loaded in the background into the example
                image = self.data_queue.get()
                self.example['image'] = image
            ready = (self.example, self.imageEpochDone)
        # get the next base example: super().next_example() calls self.get()
        self.example, self.imageEpochDone = super().next_example()
        if self.image_mode != 'no_image_model' and 'image_id' in self.example:
            # load the next image in the background
            image_id = self.example['image_id']
            self.submit_load_request(image_id)
        # Try to return the previously cached example
        if ready is None:
            return self.next_example()
        else:
            return ready

    def share(self):
        shared = super().share()
        shared['ques'] = self.ques
        if hasattr(self, 'annotation'):
            shared['annotation'] = self.annotation
        shared['image_loader'] = self.image_loader
        return shared

    def _setup_data(self, data_path, annotation_path):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.ques = json.load(data_file)

        if not self.datatype.startswith('test'):
            print('loading: ' + annotation_path)
            with PathManager.open(annotation_path) as data_file:
                self.annotation = json.load(data_file)


class McTeacher(OeTeacher):
    """
    VQA Multiple-Choice teacher, which inherits from OeTeacher but overrides the label
    and label_candidates fields with multiple choice data.
    """

    def get(self, episode_idx, entry_idx=0):
        action = super().get(episode_idx, entry_idx)
        qa = self.ques['questions'][episode_idx]
        multiple_choices = qa['multiple_choices']
        action['label_candidates'] = multiple_choices

        if not self.datatype.startswith('test'):
            anno = self.annotation['annotations'][episode_idx]
            action['labels'] = [anno['multiple_choice_answer']]

        return action


class AllTeacher(OeTeacher):
    """
    VQA Teacher, which inherits from OeTeacher and gives access to the multiple choices
    and the multiple choice answer.
    """

    def act(self):
        # parent class increments episode_idx after getting ex, so need to
        # cache the episode_idx first
        episode_idx = self.episode_idx
        action = super().act()

        qa = self.ques['questions'][episode_idx]
        multiple_choices = qa['multiple_choices']

        action['label_candidates'] = multiple_choices

        if not self.datatype.startswith('test'):
            anno = self.annotation['annotations'][episode_idx]
            self.mclabel = [anno['multiple_choice_answer']]

        if self.datatype.startswith('train'):
            action['mc_label'] = self.mclabel

        return action


class DefaultTeacher(McTeacher):
    # default to Multiple-Choice Teacher
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .regexp_tokenizer import RegexpTokenizer
from .simple_tokenizer import SimpleTokenizer


def set_default(key, value):
    global DEFAULTS
    DEFAULTS[key] = value


# Spacy is optional
try:
    from .spacy_tokenizer import SpacyTokenizer
except ImportError:
    pass


def get_class(name):
    if name == 'spacy':
        return SpacyTokenizer
    if name == 'regexp':
        return RegexpTokenizer
    if name == 'simple':
        return SimpleTokenizer

    raise RuntimeError('Invalid tokenizer: %s' % name)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Generates a controllable_gen version of a ParlAI task, i.e., for tasks with multi-turn
dialogues (episodes), this will generate a task with single example episodes, in which
we append to the context a special classification token.

In order to use this teacher, specify the task flag as follows:
`--task genderation_bias:controllable_task:<ORIGINAL TASK NAME>`.

As an example, try running:

`parlai display_data -t genderation_bias:controllable_task:convai2`
"""

from parlai.core.params import ParlaiParser
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager
import parlai.utils.logging as logging
from parlai.utils.typing import TShared

from parlai.tasks.genderation_bias.build import build
from parlai.tasks.genderation_bias.utils import (
    flatten_and_classify,
    get_original_task_module,
)

from copy import deepcopy
import datetime
import glob
import json
import os
from tqdm import tqdm
from typing import List, Optional, Tuple


class ControllableTaskTeacher(FixedDialogTeacher):
    """
    Generates a controllable_gen version of a ParlAI task, i.e., for tasks with multi-
    turn dialogues (episodes), this will generate a task with single example episodes,
    in which we append to the context a special classification token.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        flattened = parser.add_argument_group('ControllableTaskTeacher Flattening Args')
        flattened.add_argument(
            '--flatten-include-labels',
            type='bool',
            default=True,
            help='Include labels in the history when flattening an episode',
        )
        flattened.add_argument(
            '--flatten-delimiter',
            type=str,
            default='\n',
            help='How to join the dialogue history from previous turns.',
        )
        flattened.add_argument(
            '--flatten-max-context-length',
            type=int,
            default=-1,
            help='Maximum number of utterances to include per episode. '
            'Default -1 keeps all.',
        )
        agent = parser.add_argument_group('ControllableTaskTeacher Args')
        agent.add_argument(
            '--invalidate-cache',
            type='bool',
            default=False,
            help='Set this to True to rebuild the data (may want to do this if '
            'original data has changed or you want to rebuild with new options)',
        )
        agent.add_argument(
            '--max-examples',
            type=int,
            default=-1,
            help='If greater than zero, will stop building after a certain num of exs',
        )
        agent.add_argument(
            '--fixed-control',
            type=str,
            default='',
            help='Always append this fixed control string, good for deploy time.',
        )
        # Add the arguments for the task teacher
        opt = parser.parse_and_process_known_args()[0]
        tasks = get_original_task_module(opt, multi_possible=True)
        for task in tasks:
            if hasattr(task, 'add_cmdline_args'):
                task.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser

    def __init__(self, opt: Opt, shared: TShared = None):
        assert opt['flatten_delimiter'] == opt.get(
            'delimiter', '\n'
        ), '--flatten-delimiter and --delimiter are set differently, please inspect and set to the same to avoid unexpected results'
        self.opt = opt

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self.word_lists = self.build_wordlists(opt)
            self.data = self._setup_data(opt)

        super().__init__(opt, shared)
        self.reset()

    def num_episodes(self) -> int:
        return len(self.data)

    def num_examples(self) -> int:
        return len(self.data)

    def _get_save_path(self, datapath: str, date: str) -> str:
        """
        Return save path for the controllable gen data.

        :param datapath:
            path to ParlAI Data
        :param date:
            current date

        :return path:
            return path to save
        """
        return os.path.join(
            datapath,
            f"{self.original_task_name.replace(':', '_')}_flattened_controllable_gen_{date}",
        )

    @classmethod
    def build_wordlists(cls, opt: Opt) -> Tuple[List[str], List[str]]:
        """
        Load list of explicitly gendered words.

        Words taken from <https://github.com/uclanlp/gn_glove/blob/master/wordlist/>.

        Examples include brother, girl, actress, husbands, etc.
        """
        build(opt['datapath'])
        folder = os.path.join(opt['datapath'], 'genderation_bias')
        male_words = os.path.join(folder, 'male_word_file.txt')
        female_words = os.path.join(folder, 'female_word_file.txt')

        with open(male_words, 'r') as f:
            male = f.read().splitlines()

        with open(female_words, 'r') as f:
            female = f.read().splitlines()

        return male, female

    def _setup_data(self, opt: Opt) -> List[List[Message]]:
        """
        Flatten and classify the normal task data.

        Save/load where applicable.

        :param opt:
            options dict.
        """
        # create save directory, if it does not already exist
        self.original_task_name = ':'.join(opt['task'].split(':')[2:])
        self.save_dir = self._get_save_path(
            opt['datapath'], str(datetime.datetime.today())
        )
        os.makedirs(self.save_dir, exist_ok=True)

        fname = f"{opt['datatype'].split(':')[0]}.json"
        self.save_path = os.path.join(self.save_dir, fname)

        data = self.load_data(opt, fname)
        if data is not None:
            # successfully load data
            return data

        # build the original teacher
        original_task_module = get_original_task_module(opt)
        teacher_opt = deepcopy(opt)
        teacher_opt['task'] = self.original_task_name
        teacher = original_task_module(teacher_opt)

        total_exs = teacher.num_examples()
        if self.opt['max_examples'] > 0:
            total_exs = min(self.opt['max_examples'], total_exs)

        progress_bar = tqdm(
            total=total_exs, unit='ex', unit_scale=True, desc='Building flattened data'
        )

        all_episodes = []
        num_exs = 0
        while num_exs < total_exs:
            current_episode = []
            episode_done = False

            while not episode_done:
                action = Message(teacher.act())
                current_episode.append(action)
                episode_done = action.get('episode_done', False)
                num_exs += 1

            # flatten the episode into 1-example episodes with context
            flattened_ep = flatten_and_classify(
                current_episode,
                opt['flatten_max_context_length'],
                include_labels=opt['flatten_include_labels'],
                delimiter=opt['flatten_delimiter'],
                word_lists=self.word_lists,
            )
            all_episodes += flattened_ep

            progress_bar.update(len(flattened_ep))

        # save data for future use
        self.save_data(all_episodes)

        return all_episodes

    def load_data(self, opt: Opt, filename: str) -> Optional[List[List[Message]]]:
        """
        Attempt to load pre-build data.

        Checks for the most recently build data via the date string.

        :param opt:
            options dict
        :param filename:
            name of (potentially) saved data

        :return episodes:
            return list of episodes, if available
        """
        # first check for the most recent date
        save_dir = self._get_save_path(opt['datapath'], '*')
        all_dates = []
        for fname in glob.glob(os.path.join(save_dir, filename)):
            date = os.path.split(fname)[0].split('_')[-1]
            all_dates.append(date)

        if len(all_dates) > 0:
            most_recent = os.path.join(
                self._get_save_path(opt['datapath'], sorted(all_dates)[-1]), filename
            )
        else:
            # data has not been built yet
            return None

        if opt['invalidate_cache']:
            # invalidate the cache and remove the existing data
            logging.warn(
                f' [ WARNING: invalidating cache at {self.save_path} and rebuilding the data. ]'
            )
            if self.save_path == most_recent:
                os.remove(self.save_path)
            return None

        # Loading from most recent date
        self.save_path = most_recent
        logging.info(f' [ Data already exists. Loading from: {self.save_path} ]')
        with PathManager.open(self.save_path, 'rb') as f:
            data = json.load(f)

        return data

    def save_data(self, data: List[List[Message]]):
        """
        Save the data via dumping to a json file.

        :param data:
            list of episodes
        """
        try:
            json_data = json.dumps(data)
            with PathManager.open(self.save_path, 'w') as f:
                f.write(json_data)
            logging.info(f'[ Data successfully saved to path: {self.save_path} ]')
        except Exception:
            logging.warn('Data is not json serializable; not saving')

    def get(self, episode_idx: int, entry_idx: int = 0) -> Message:
        """
        Return a flattened example.

        If using a fixed control, put that in instead of what was originally in the text.

        :param episode_idx:
            index of ep in data
        :param entry_idx:
            index of ex in ep

        :return ex:
            return an example
        """
        ex = Message(self.data[episode_idx])

        if self.opt['fixed_control'] != '':
            old_text = ' '.join(ex['text'].split(' ')[:-1])
            text = f"{old_text} {self.opt['fixed_control']}"
            ex.force_set('text', text)

        return ex

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared


class DefaultTeacher(ControllableTaskTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Utils for Controllable Gen Teacher.
"""
from collections import deque
import random

from parlai.core.loader import load_teacher_module
from parlai.core.message import Message
from parlai.core.opt import Opt

from typing import List, Tuple

PUNCTUATION_LST = [
    (' .', '.'),
    (' !', '!'),
    (' ?', '?'),
    (' ,', ','),
    (" ' ", "'"),
    (" . . . ", "... "),
    (" ( ", " ("),
    (" ) ", ") "),
    (" ; ", "; "),
]


def format_text(text: str, lower: bool = True) -> str:
    """
    Space punctuation and lowercase text.

    :param text:
        text to lowercase
    :param lower:
        whether to lowercase or not

    :return text:
        return formatted text.
    """
    if lower:
        text = text.lower()
    for punc in PUNCTUATION_LST:
        text = text.replace(punc[1], punc[0])

    return text


def get_word_list_token(text: str, word_lists: Tuple[List[str], List[str]]) -> str:
    """
    Return a control token corresponding to gender within text.

    :param text:
        text to consider for control token
    :param word_lists:
        tuple of lists for male-specific and female-specific words

    :return token:
        return control token corresponding to input text.
    """
    m_list, f_list = word_lists
    text = format_text(text)
    m_cnt = 0
    f_cnt = 0
    for word in text.split(' '):
        if word in m_list:
            m_cnt += 1
        if word in f_list:
            f_cnt += 1

    if f_cnt == 0 and m_cnt == 0:
        return 'f0m0'
    elif f_cnt == 0 and m_cnt > 0:
        return 'f0m1'
    elif f_cnt > 0 and m_cnt == 0:
        return 'f1m0'
    else:
        return 'f1m1'


def flatten_and_classify(
    episode: List[Message],
    context_length: int,
    word_lists: Tuple[List[str], List[str]],
    include_labels: bool = True,
    delimiter: str = '\n',
):
    """
    Flatten the dialogue history of an episode, explode into N new examples.

    Additionally, add control token corresponding to gender identified in the
    episode.

    :param episode:
        list of examples to flatten
    :param context_length:
        max number of utterances to use while flattening
    :param word_lists:
        tuple of lists for male-specific and female-specific words
    :param include_labels:
        whether to include labels while flattening
    :param delimiter:
        delimiter to use while flattening
    """
    context = deque(maxlen=context_length if context_length > 0 else None)
    new_episode = []

    for ex in episode:
        context.append(ex.get('text', ''))
        # add context
        if len(context) > 1:
            ex.force_set('text', delimiter.join(context))
        # set episode_done to be True
        ex.force_set('episode_done', True)
        labels = ex.get('labels', ex.get('eval_labels', None))
        if labels is not None and include_labels:
            context.append(random.choice(labels))

        # word list
        control_tok = get_word_list_token(random.choice(labels), word_lists)
        ex.force_set('text', ex['text'] + ' ' + control_tok)
        new_episode.append(ex)

    return new_episode


def get_original_task_module(opt: Opt, multi_possible: bool = False):
    """
    Returns task module of "original" task.

    Original task in this case means the task we want to use
    with the control teacher.

    :param opt:
        opt dict
    :param multi_possible:
        specify True if multiple tasks are possible.

    :return task_module:
        return module associated with task.
    """
    modules = []
    tasks = opt['task'].split(',')
    if not multi_possible:
        assert len(tasks) == 1

    for task in tasks:
        if len(task.split(':')) < 3:
            raise RuntimeError(
                '\n\n********************************************************\n'
                'Must specify original task using the following format:\n'
                '`--task internal:flattened:task:<ORIGINAL TASK NAME>`'
                '\n********************************************************\n'
            )
        original_task = ':'.join(task.split(':')[2:])
        task_module = load_teacher_module(original_task)
        modules.append(task_module)

    if multi_possible:
        return modules

    return modules[0]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.mutators import register_mutator, MessageMutator


@register_mutator("word_reverse")
class WordReverseMutator(MessageMutator):
    """
    Reverses the words of each turn.

    Only the text (prompt) is modified, not the labels. Utterances separated by newlines
    will not be shuffled across boundaries. You may wish to combine it with the flatten
    mutator to reverse both labels and texts.
    """

    def __init__(self, opt: Opt):
        super().__init__(opt)
        self.rng = random.Random(42)

    def message_mutation(self, message: Message) -> Message:
        texts = message.pop('text').split('\n')
        output_texts = []
        for text in texts:
            words = text.split(' ')
            words = list(reversed(words))
            output_texts.append(' '.join(words))
        message['text'] = '\n'.join(output_texts)
        return message



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.io import PathManager
from parlai.core.teachers import FixedDialogTeacher
from parlai.core.image_featurizers import ImageLoader
from .build import build
from parlai.tasks.coco_caption.build_2014 import buildImage as buildImage_2014
from parlai.tasks.coco_caption.build_2015 import buildImage as buildImage_2015

import json
import os


def _path(opt):
    build(opt)
    buildImage_2014(opt)
    buildImage_2015(opt)
    dt = opt['datatype'].split(':')[0]

    img_version = None
    if dt == 'train':
        ques_suffix = 'v2_OpenEnded_mscoco_train2014'
        annotation_suffix = 'v2_mscoco_train2014'
        img_suffix = os.path.join('train2014', 'COCO_train2014_')
        img_version = '2014'
    elif dt == 'valid':
        ques_suffix = 'v2_OpenEnded_mscoco_val2014'
        annotation_suffix = 'v2_mscoco_val2014'
        img_suffix = os.path.join('val2014', 'COCO_val2014_')
        img_version = '2014'
    elif dt == 'test':
        ques_suffix = 'v2_OpenEnded_mscoco_test2015'
        annotation_suffix = 'None'
        img_suffix = os.path.join('test2015', 'COCO_test2015_')
        img_version = '2015'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], 'VQA-v2', ques_suffix + '_questions.json')

    annotation_path = os.path.join(
        opt['datapath'], 'VQA-v2', annotation_suffix + '_annotations.json'
    )

    image_path = os.path.join(
        opt['datapath'], 'COCO-IMG-{}'.format(img_version), img_suffix
    )

    return data_path, annotation_path, image_path


class OeTeacher(FixedDialogTeacher):
    """
    VQA v2.0 Open-Ended teacher, which loads the json VQA data and implements the
    ``get`` method to return additional metadata.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt)
        self.image_mode = opt.get('image_mode', 'no_image_model')

        if shared and 'ques' in shared:
            # another instance was set up already, just reference its data
            self.ques = shared['ques']
            if 'annotation' in shared:
                self.annotation = shared['annotation']
            self.image_loader = shared['image_loader']
        else:
            # need to set up data from scratch
            data_path, annotation_path, self.image_path = _path(opt)
            self._setup_data(data_path, annotation_path)
            self.image_loader = ImageLoader(opt)

        self.reset()

    def reset(self):
        super().reset()  # call parent reset so other fields can be set up
        self.example = None  # set up caching fields

    def num_examples(self):
        return len(self.ques['questions'])

    def num_episodes(self):
        return self.num_examples()

    def submit_load_request(self, image_id):
        img_path = self.image_path + '%012d.jpg' % (image_id)
        self.data_loader.request_load(
            self.receive_data, self.image_loader.load, (img_path,)
        )

    def get(self, episode_idx, entry_idx=0):
        qa = self.ques['questions'][episode_idx]
        question = qa['question']

        action = {'text': question, 'image_id': qa['image_id'], 'episode_done': True}

        if not self.datatype.startswith('test'):
            # test set annotations are not available for this dataset
            anno = self.annotation['annotations'][episode_idx]
            action['labels'] = [ans['answer'] for ans in anno['answers']]

        return action

    def next_example(self):
        """
        Returns the next example from this dataset after starting to queue up the next
        example.
        """
        ready = None
        # pull up the currently queued example
        if self.example is not None:
            if self.image_mode != 'no_image_model':
                # move the image we loaded in the background into the example
                image = self.data_queue.get()
                self.example['image'] = image
            ready = (self.example, self.epochDone)
        # get the next base example: super().next_example() calls self.get()
        self.example, self.epochDone = super().next_example()
        if self.image_mode != 'no_image_model' and 'image_id' in self.example:
            # load the next image in the background
            image_id = self.example['image_id']
            self.submit_load_request(image_id)
        # Try to return the previously cached example
        if ready is None:
            return self.next_example()
        else:
            return ready

    def share(self):
        shared = super().share()
        shared['ques'] = self.ques
        if hasattr(self, 'annotation'):
            shared['annotation'] = self.annotation
        shared['image_loader'] = self.image_loader
        return shared

    def _setup_data(self, data_path, annotation_path):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            self.ques = json.load(data_file)

        if not self.datatype.startswith('test'):
            print('loading: ' + annotation_path)
            with PathManager.open(annotation_path) as data_file:
                self.annotation = json.load(data_file)


class AllTeacher(OeTeacher):
    """
    VQA v2.0 Open-Ended teacher, which inherits from OeTeacher and gives access to the
    multiple choice answer.
    """

    def act(self):
        action = super().act()

        if not self.datatype.startswith('test'):
            anno = self.annotation['annotations'][self.episode_idx]
            self.mclabel = [anno['multiple_choice_answer']]

        if self.datatype.startswith('train'):
            action['mc_label'] = self.mclabel

        return action


class DefaultTeacher(OeTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Creates supplementary training examples using a trained model and unused train file.

Input:
    model: a trained agent
    deploy: an unused train file (labels available)
    flags: (settings)
Returns:
    supp: a file in parlai_meta format with new training examples


Examples
--------

.. code-block:: shell

  python create_synth_supp.py
  --model-file /path/to/model-file.model
  --deploy-file /path/to/deploy-file.txt
  --supp-file /path/to/supp-file.txt
  --conversion-rate 0.9
  --conversion-acc 0.9
"""
import json
import os
import random

from parlai.core.agents import create_agent
from parlai.core.params import ParlaiParser
from parlai.core.worlds import create_task
from parlai.utils.io import PathManager

from projects.self_feeding.utils import Parley

PARLAIHOME = os.environ['PARLAIHOME']
NUM_INLINE_CANDS = 20


def setup_args(parser=None):
    # Get command line arguments
    if parser is None:
        parser = ParlaiParser(True, True, 'Display data from a task')
    parser.add_argument(
        '-mf',
        '--model-file',
        type=str,
        default='/private/home/bhancock/metadialog/models/unknown.mdl',
    )
    parser.add_argument('-dfile', '--deploy-file', type=str, default='train_c')
    parser.add_argument('-sfile', '--supp-file', type=str, default='supp_c')
    parser.add_argument(
        '-cr',
        '--conversion-rate',
        type=float,
        default=1.0,
        help="The fraction of misses converted into new training data",
    )
    parser.add_argument(
        '-ca',
        '--conversion-acc',
        type=float,
        default=1.0,
        help="The fraction of converted data that have a correct label",
    )
    parser.set_defaults(bs=1)
    parser.set_defaults(ecands='inline')
    parser.set_defaults(datatype='valid')
    return parser


def create_supp(opt):
    """
    Evaluates a model.

    :param opt: tells the evaluation function how to run
    :return: the final result of calling report()
    """
    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    world = create_task(opt, agent)

    # Extract supp examples from misses on deploy set
    num_seen = 0
    num_misses = 0
    num_supp = 0
    num_supp_correct = 0
    examples = []
    while not world.epoch_done():
        world.parley()
        # Examples are considered one at a time
        num_seen += 1
        if num_seen % 1000 == 0:
            print(f"{num_seen}/{world.num_examples()}")
        report = world.report()
        if report['accuracy'] < 1.0:
            # Example is a miss (i.e., model got it wrong)
            num_misses += 1
            if random.random() < opt['conversion_rate']:
                # Example will be converted (e.g., bot recognized mistake and asked)
                num_supp += 1
                texts = world.acts[0]['text'].split('\n')
                context = texts[-1]
                memories = texts[:-1]
                candidates = world.acts[0]['label_candidates']
                # Reward of 1 indicates positive, -1 indicates negative (for training)
                # For now, we only train with positives, and the reward field is unused
                reward = 1

                if random.random() < opt['conversion_acc']:
                    # Example will be converted correctly (e.g., good user response)
                    num_supp_correct += 1
                    response = world.acts[0]['eval_labels'][0]
                else:
                    # Example will be converted incorrectly (e.g., bad user response)
                    response = random.choice(
                        world.acts[0]['label_candidates'][: NUM_INLINE_CANDS - 1]
                    )

                example = Parley(context, response, reward, candidates, memories)
                examples.append(example)
        world.reset_metrics()

    print("EPOCH DONE")
    print(f"Model file: {opt['model_file']}")
    print(f"Deploy file: {opt['task']}")
    print(f"Supp file: {opt['outfile']}")
    print(f"Deploy size (# examples seen): {num_seen}")
    print(f"Supp size (# examples converted): {num_supp}")

    acc = 1 - (num_misses / num_seen)
    print(f"Accuracy (% of deploy): {acc * 100:.1f}% ({num_misses} misses)")
    print(
        f"Conversion rate (% of misses): {num_supp/num_misses * 100:.2f}% "
        f"({num_supp}/{num_misses})"
    )
    print(
        f"Conversion acc (% of converted): {num_supp_correct/num_supp * 100:.2f}% "
        f"({num_supp_correct}/{num_supp})"
    )

    with PathManager.open(opt['outfile'], 'w') as outfile:
        for ex in examples:
            outfile.write(json.dumps(ex.to_dict()) + '\n')


if __name__ == '__main__':
    random.seed(42)
    parser = setup_args()
    opt = parser.parse_args()
    opt['dict_file'] = os.path.splitext(opt['model_file'])[0] + '.dict'
    opt['task'] = 'self_feeding:convai2:' + opt['deploy_file']
    opt['outfile'] = PARLAIHOME + '/data/convai2meta/' + opt['supp_file'] + '.txt'
    create_supp(opt)
